/* ########################################################################
    Copyright (c) 1997	Bruker Analytical X-ray Systems
			6300 Enterprise Lane
			Madison, WI 53719-1173 USA
//License:
//--------
//Except for the unwarp and spatial algorithm (for which the source files have
//been replaced by the object library: unwarp.lib) and the sources provided by
//the Open JPEG Foundation (see folder jpeg), all the sources contained in this
//ZIP distribution are non-proprietary information. For example, Bruker AXS Inc.
//considers all Bruker data file formats to be non-proprietary information. You
//are free to use the Bruker supplied public domain sources in any way you see
//fit.

*+ ########################################################################
FRM_LIB: Routines for reading, writing, and handling frame files.

    To write a FRM file:
        frm_open_file
        frm_write_header
        frm_write_data
        frm_close_file
*-
   Version 1.000 -- KLS -- 20May97
*/
/* Include files */
#include "FRM_LIB.H"    /* specific to FRM files */
#include <fcntl.h>      /* _open, oflags, etc. */
#include <stdio.h>      /* printf, etc. */
#include <stdlib.h>     /* __min, etc. */
#ifdef _MSC_VER
#include "io.h"         /* _open, etc. */
#else
#define strnicmp strncasecmp
#endif
#include "errno.h"
#include "string.h"     /* _strnicmp, etc. */
#include "sys/types.h"  /* _open, etc. */
#include "sys/stat.h"   /* _open, etc. */

#ifndef _MSC_VER
int min(int a, int b)
	{
	if(a>b)
		return b;
	else
		return a;
	}
#endif

char* strupr(char* src)
	{
	char* dummy;
	int i;
	dummy=(char*) malloc(strlen(src));
	strcpy(dummy,src);
	for(i=0;i<strlen(dummy);i++)
		dummy[i]=toupper(dummy[i]);
	return dummy;
	}

long dbglevel()
/**-*/
{
    char *str = NULL;
    static long level=-1;       /* -1= routine has not yet been called */

    /* subsequent calls to dbglevel simple return debug level */
    if (level != -1) return level;

    /* get current debug level from environment variable */
    str = getenv("SAXI$DBGLEVEL");
    if (str == NULL) str = getenv("SAXI_DBGLEVEL");
    if (str != NULL) level = atol(str);
    else level=0;               /* default is no output level */
    return level;
}	
	
/**+ -----------------------------------------------------------------------
FRM_OPEN_FILE: Either opens an old FRM file or creates a new FRM file.
      Verifies that old files are indeed a FRM file.
*/
int  frm_open_file(             /* (W) File handle or -1 on error */
char* frmfilename,              /* (R) Filename for FRM file */
long itype,                     /* (R) 0=open old, !0= create new file */
long* nformat,                  /* (W) 86 (old) or 100 (new) */
long* ier)                      /* (W) Error code: 0=none, -1=not FRM file */
/**-*/
{
    char id[88];
    int  fh;                    /* file handle */
#ifdef _MSC_VER
	int  oflags = O_RDONLY | O_BINARY;
#else
	int  oflags = O_RDONLY;
#endif
    int  sflags = S_IREAD | S_IWRITE;

    /* Try to open the requested file */
#ifdef _MSC_VER
    if (itype) oflags = O_RDWR | O_BINARY | O_CREAT;
#else
	if (itype) oflags = O_RDWR | O_CREAT;
#endif
    *ier = 0;
    fh = open(frmfilename, oflags, sflags);
    if (fh == -1) {*ier = errno; goto BAGIT;}

    /* Verify that an OLD file is actually a FRM file */
    if (itype == 0) {
        /* First 8 bytes are "FORMAT :" */
        if (dbglevel() > 3) printf("\nCheck for BRUKER frame file format\n");
        read(fh, &id, sizeof(id));
        if (id[0] != 'F') *ier = -1;
        if (id[1] != 'O') *ier = -1;
        if (id[2] != 'R') *ier = -1;
        if (id[3] != 'M') *ier = -1;
        if (id[4] != 'A') *ier = -1;
        if (id[5] != 'T') *ier = -1;
        if (id[6] != ' ') *ier = -1;
        if (id[7] != ':') *ier = -1;
        /* 81 to 88 bytes are "VERSION:" */
        if (id[80] != 'V') *ier = -1;
        if (id[81] != 'E') *ier = -1;
        if (id[82] != 'R') *ier = -1;
        if (id[83] != 'S') *ier = -1;
        if (id[84] != 'I') *ier = -1;
        if (id[85] != 'O') *ier = -1;
        if (id[86] != 'N') *ier = -1;
        if (id[87] != ':') *ier = -1;
        if (*ier == -1) {close(fh); return -1;}

        /* Old format or PROTEUM format? */
        sscanf(id+8, "%d", nformat);
        //if (id[8] == '1') {
        //    printf("\nNote: FRM2FRM cannot read the new PROTEUM frame formats\n");
        //    printf(  "      Use PROTEUM to convert to old Bruker frame formats\n");
        //    _close(fh);
        //    return -1;
        //}

        /* Restore file pointer to BOF */
        if (lseek(fh, 0, 0) == -1) *ier = errno;
    }

    /* Exit with error code */
BAGIT:
    return (fh);
}

/**+ -----------------------------------------------------------------------
FRM_CLOSE_FILE: Closes a FRM file opened with "frm_open_file".
*/
long frm_close_file(            /* (W) Error code: 0=none */
int  fhFRM)                     /* (R) File handle (from frm_open_file) */
/**-*/
{
    long ier = 0;               /* Error code: 0=none */

    /* Try to close the requested file */
    if (close(fhFRM)) ier = errno;
    return ier;
}

/**+ -----------------------------------------------------------------------
FRM_WRITE_HEADER: Writes initial FRM file header. Assumes file pointer is
      at BOF, which is true for newly created files.
*/
long frm_write_header(          /* (W) Error code: 0=none */
int  fhFRM,                     /* (R) File handle (from frm_open_file) */
struct FRM_HEADER *frm,         /* (R) FRM header information */
long* nbytes)                   /* (W) Number of bytes written, -1=error */
/**-*/
{
    long ier = 0;               /* Error code: 0=none */

    /* Check frame header for completeness? n.y.i. */

    /* Try to write header into FRM file */
    *nbytes = write(fhFRM, frm, sizeof(struct FRM_HEADER));
    if (*nbytes == -1) ier = errno;
    return ier;
}

/**+ -----------------------------------------------------------------------
FRM_WRITE_DATA: Writes image data to FRM file. Assumes the
      file pointer is after the frame header.
*/
long frm_write_data(            /* (W) Error code: 0=none */
int  fhFRM,                     /* (R) File handle (from frm_open_file) */
long ncounts,                   /* (R) Number of counts */
long nbytespercount,            /* (R) Number of bytes per count: 1, 2, 4 */
long counts[],                  /* (R) Image counts */
long* nbytes)                   /* (W) Number of bytes written, -1=error */
/**-*/
{
    long ier = 0;               /* Error code: 0=none */

    /* Try to write image data into FRM file */
    *nbytes = write(fhFRM, counts, (int) (ncounts * nbytespercount));
    if (*nbytes == -1) {ier = errno; goto BAGIT;}

    /* Exit with error code */
BAGIT:
    return ier;
}

/**+ -----------------------------------------------------------------------
FRM_WRITE_OVERFLOW: Writes overflow table from FRM file. Assumes the
      file pointer is after the frame image.
*/
long frm_write_overflow(        /* (W) Error code: 0=none */
int  fhFRM,                     /* (R) File handle (from frm_open_file) */
long nsize,                     /* (R) Number of bytes in table (min=512) */
char *table,                    /* (W) Overflow table */
long* nbytes)                   /* (W) Number of bytes read, -1=error */
/**-*/
{
    long ier = 0;               /* Error code: 0=none */

    /* Try to write overflow table from FRM file */
    *nbytes = write(fhFRM, table, (int) nsize);
    if (*nbytes == -1) {ier = errno; goto BAGIT;}

    /* Exit with error code */
BAGIT:
    return ier;
}

/**+ -----------------------------------------------------------------------
FRM_READ_HEADER: Reads initial FRM file header. Assumes file pointer is
      at BOF, which is true for newly created files.

      Note: Also cleans-up known problems with FRM headers. (n.y.i.)
*/
long frm_read_header(           /* (W) Error code: 0=none */
int  fhFRM,                     /* (R) File handle (from frm_open_file) */
struct FRM_HEADER *frm,         /* (W) FRM header information */
long* nbytes)                   /* (W) Number of bytes read, -1=error */
/**-*/
{
    long ier = 0;               /* Error code: 0=none */

    /* Try to read header from FRM file */
    *nbytes = read(fhFRM, frm, sizeof(struct FRM_HEADER));
    if (*nbytes == -1) {ier = errno; goto BAGIT;}

    /* Clean-up known problems in FRM headers */
    /* n.y.i. */

BAGIT:
    return ier;
}

/**+ -----------------------------------------------------------------------
FRM_READ_DATA: Reads image data from FRM file. Assumes the
      file pointer is after the frame header.
*/
long frm_read_data(             /* (W) Error code: 0=none */
int  fhFRM,                     /* (R) File handle (from frm_open_file) */
long ncounts,                   /* (R) Number of counts */
long nbytespercount,            /* (R) Number of bytes per count: 1, 2, 4 */
long counts[],                  /* (W) Image counts */
long* nbytes)                   /* (W) Number of bytes read, -1=error */
/**-*/
{
    long ier = 0;               /* Error code: 0=none */

    /* Try to read image data from FRM file */
    *nbytes = read(fhFRM, counts, (int) (ncounts * nbytespercount));
    if (*nbytes == -1) {ier = errno; goto BAGIT;}

    /* Exit with error code */
BAGIT:
    return ier;
}

/**+ -----------------------------------------------------------------------
FRM_READ_OVERFLOW: Reads overflow table from FRM file. Assumes the
      file pointer is after the frame image.
*/
long frm_read_overflow(         /* (W) Error code: 0=none */
int  fhFRM,                     /* (R) File handle (from frm_open_file) */
long nsize,                     /* (R) Number of bytes in table */
char *table,                    /* (W) Overflow table */
long* nbytes)                   /* (W) Number of bytes read, -1=error */
/**-*/
{
    long ier = 0;               /* Error code: 0=none */

    /* Try to read overflow table from FRM file */
    *nbytes = read(fhFRM, table, (int) nsize);
    if (*nbytes == -1) {ier = errno; goto BAGIT;}

    /* Exit with error code */
BAGIT:
    return ier;
}

/**+ -----------------------------------------------------------------------
FRM_MODIFY_HEADER: Modify a line in the frame header structure.
*/
long frm_modify_header(         /* (W) Error code: 0=none */
char *hdrline,                  /* (W) */
char *newline)                  /* (R) */
/**-*/
{
    // Determine length of string (to \0, \n, or \r characters)
    int n;
    long num = strlen(newline); // to \0 character
    for (n=num-1; n>=0; n--) {
        if (newline[n] == '\n')
            num = n;
        if (newline[n] == '\r')
            num = n;
    }

    // Copy only the text to header line
    memset(hdrline, ' ', 80);
    if (num > 0)
        strncpy(hdrline, newline, min(80, num));
    return 0;
}

/**+ -----------------------------------------------------------------------
FRM_DEFAULT_HEADER: Initialize FRM HEADER structure with default values.
    Unfortunately, one cannot default by clearing memory to zero, thus
    the need for this routine.

    Note: Does not default wavelengths. Defaults to VERSION ?? format.
*/
//long frm_default_header(        /* (W) Error code: 0=none */
//struct FRM_HEADER *frm)         /* (W) FRM file header information */
/**-*/
//{
//    extern long frm_default_header_usersettings(struct FRM_HEADER *frm);
//    char newline[81];

    /* default file header: everything, except wavelengths ------------- */
  /*  memset(frm, '.', sizeof(struct FRM_HEADER));
    strcpy(newline, "FORMAT :        86");
    frm_modify_header(frm->line[0], newline);
    strcpy(newline, "VERSION:         8");
    frm_modify_header(frm->line[1], newline);
    strcpy(newline, "HDRBLKS:        15");
    frm_modify_header(frm->line[2], newline);
    strcpy(newline, "TYPE   :DEFAULT");
    frm_modify_header(frm->line[3], newline);
    strcpy(newline, "SITE   :Unknown");
    frm_modify_header(frm->line[4], newline);
    strcpy(newline, "MODEL  :Unknown");
    frm_modify_header(frm->line[5], newline);
    strcpy(newline, "USER   :Unknown");
    frm_modify_header(frm->line[6], newline);
    strcpy(newline, "SAMPLE :Unknown");
    frm_modify_header(frm->line[7], newline);
    strcpy(newline, "SETNAME:Unknown");
    frm_modify_header(frm->line[8], newline);
    strcpy(newline, "RUN    :         0");
    frm_modify_header(frm->line[9], newline);
    strcpy(newline, "SAMPNUM:         0");
    frm_modify_header(frm->line[10], newline);
    strcpy(newline, "TITLE  :");
    frm_modify_header(frm->line[11], newline);
    frm_modify_header(frm->line[12], newline);
    frm_modify_header(frm->line[13], newline);
    frm_modify_header(frm->line[14], newline);
    frm_modify_header(frm->line[15], newline);
    frm_modify_header(frm->line[16], newline);
    frm_modify_header(frm->line[17], newline);
    frm_modify_header(frm->line[18], newline);
    strcpy(newline, "NCOUNTS:         0");
    frm_modify_header(frm->line[19], newline);
    strcpy(newline, "NOVERFL:         0");
    frm_modify_header(frm->line[20], newline);
    strcpy(newline, "MINIMUM:         0");
    frm_modify_header(frm->line[21], newline);
    strcpy(newline, "MAXIMUM:       255");
    frm_modify_header(frm->line[22], newline);
    strcpy(newline, "NONTIME:         0");
    frm_modify_header(frm->line[23], newline);
    strcpy(newline, "NLATE  :         0");
    frm_modify_header(frm->line[24], newline);
    strcpy(newline, "FILENAM:Unknown");     // n.y.i.
    frm_modify_header(frm->line[25], newline);
    strcpy(newline, "CREATED:");            // n.y.i.
    frm_modify_header(frm->line[26], newline);
    strcpy(newline, "CUMULAT:     0.0000000");
    frm_modify_header(frm->line[27], newline);
    strcpy(newline, "ELAPSDR:     0.0000000");
    frm_modify_header(frm->line[28], newline);
    strcpy(newline, "ELAPSDA:     0.0000000");
    frm_modify_header(frm->line[29], newline);
    strcpy(newline, "OSCILLA:         0");
    frm_modify_header(frm->line[30], newline);
    strcpy(newline, "NSTEPS :         1");
    frm_modify_header(frm->line[31], newline);
    strcpy(newline, "RANGE  :     0.2500000");
    frm_modify_header(frm->line[32], newline);
    strcpy(newline, "START  :     0.0000000");
    frm_modify_header(frm->line[33], newline);
    strcpy(newline, "INCREME:     0.2500000");
    frm_modify_header(frm->line[34], newline);
    strcpy(newline, "NUMBER :         0");
    frm_modify_header(frm->line[35], newline);
    strcpy(newline, "NFRAMES:         1");
    frm_modify_header(frm->line[36], newline);
    strcpy(newline, "ANGLES :     0.0000000     0.0000000     0.0000000     0.0000000");
    frm_modify_header(frm->line[37], newline);
    strcpy(newline, "NOVER64:         0");
    frm_modify_header(frm->line[38], newline);
    strcpy(newline, "NPIXELB:         1");
    frm_modify_header(frm->line[39], newline);
    strcpy(newline, "NROWS  :       512");
    frm_modify_header(frm->line[40], newline);
    strcpy(newline, "NCOLS  :       512");
    frm_modify_header(frm->line[41], newline);
    strcpy(newline, "WORDORD:         0");
    frm_modify_header(frm->line[42], newline);
    strcpy(newline, "LONGORD:         0");
    frm_modify_header(frm->line[43], newline);
    strcpy(newline, "TARGET :CU");
    frm_modify_header(frm->line[44], newline);
    strcpy(newline, "SOURCEK:    40.0000000");
    frm_modify_header(frm->line[45], newline);
    strcpy(newline, "SOURCEM:    40.0000000");
    frm_modify_header(frm->line[46], newline);
    strcpy(newline, "FILTER :Unknown");
    frm_modify_header(frm->line[47], newline);
    strcpy(newline, "CELL   :     0.0000000     0.0000000     0.0000000     0.0000000     0.0000000");
    frm_modify_header(frm->line[48], newline);
    strcpy(newline, "CELL   :     0.0000000");
    frm_modify_header(frm->line[49], newline);
    strcpy(newline, "MATRIX :     0.0000000     0.0000000     0.0000000     0.0000000     0.0000000");
    frm_modify_header(frm->line[50], newline);
    strcpy(newline, "MATRIX :     0.0000000     0.0000000     0.0000000     0.0000000");
    frm_modify_header(frm->line[51], newline);
    strcpy(newline, "LOWTEMP:         0");
    frm_modify_header(frm->line[52], newline);
    strcpy(newline, "ZOOM   :     0.5000000     0.5000000     1.0000000");
    frm_modify_header(frm->line[53], newline);
    strcpy(newline, "CENTER :     0.0000000     0.0000000");
    frm_modify_header(frm->line[54], newline);
    strcpy(newline, "DISTANC:    12.0000000");
    frm_modify_header(frm->line[55], newline);
    strcpy(newline, "TRAILER:        -1");
    frm_modify_header(frm->line[56], newline);
    strcpy(newline, "COMPRES:NONE");
    frm_modify_header(frm->line[57], newline);
    strcpy(newline, "LINEAR :     1.0000000     0.0000000");
    frm_modify_header(frm->line[58], newline);
    strcpy(newline, "PHD    :     0.0000000     0.0000000");
    frm_modify_header(frm->line[59], newline);
    strcpy(newline, "PREAMP :     0.0000000");
    frm_modify_header(frm->line[60], newline);
    strcpy(newline, "CORRECT:Unknown");
    frm_modify_header(frm->line[61], newline);
    strcpy(newline, "WARPFIL:Unknown");
    frm_modify_header(frm->line[62], newline);
    strcpy(newline, "WAVELEN:     1.5418400     1.5405600      1.5443900");
    frm_modify_header(frm->line[63], newline);
    strcpy(newline, "MAXXY  :   255.0000000   255.0000000");
    frm_modify_header(frm->line[64], newline);
    strcpy(newline, "AXIS   :         2");
    frm_modify_header(frm->line[65], newline);
    strcpy(newline, "ENDING :     0.0000000     0.0000000     0.0000000     0.0000000");
    frm_modify_header(frm->line[66], newline);
    strcpy(newline, "DETPAR :     0.0000000     0.0000000     0.0000000     0.0000000     0.0000000");
    frm_modify_header(frm->line[67], newline);
    strcpy(newline, "DETPAR :     0.0000000");
    frm_modify_header(frm->line[68], newline);
    strcpy(newline, "LUT    :NONE");
    frm_modify_header(frm->line[69], newline);
    strcpy(newline, "DISPLIM:     0.0000000     0.0000000");
    frm_modify_header(frm->line[70], newline);
    strcpy(newline, "PROGRAM:Unknown");
    frm_modify_header(frm->line[71], newline);
    strcpy(newline, "ROTATE :         0");
    frm_modify_header(frm->line[72], newline);
    strcpy(newline, "BITMASK:$NULL");
    frm_modify_header(frm->line[73], newline);
    strcpy(newline, "OCTMASK:           0          60           0         451         511         962");
    frm_modify_header(frm->line[74], newline);
    strcpy(newline, "OCTMASK:         511         451");
    frm_modify_header(frm->line[75], newline);
    strcpy(newline, "ESDCELL:     0.0000000     0.0000000     0.0000000     0.0000000     0.0000000");
    frm_modify_header(frm->line[76], newline);
    strcpy(newline, "ESDCELL:     0.0000000");
    frm_modify_header(frm->line[77], newline);
    strcpy(newline, "DETTYPE:Unknown");
    frm_modify_header(frm->line[78], newline);
    strcpy(newline, "NEXP   :         0         0");
    frm_modify_header(frm->line[79], newline);
    strcpy(newline, "CCDPARM:     0.0000000     0.0000000     0.0000000     0.0000000     0.0000000");
    frm_modify_header(frm->line[80], newline);
    strcpy(newline, "CHEM   :?");
    frm_modify_header(frm->line[81], newline);
    strcpy(newline, "MORPH  :?");
    frm_modify_header(frm->line[82], newline);
    strcpy(newline, "CCOLOR :?");
    frm_modify_header(frm->line[83], newline);
    strcpy(newline, "CSIZE  :?|?|?|?|?");
    frm_modify_header(frm->line[84], newline);
    strcpy(newline, "DNSMET :?");
    frm_modify_header(frm->line[85], newline);
    strcpy(newline, "DARK   :Unknown");
    frm_modify_header(frm->line[86], newline);
    strcpy(newline, "AUTORNG:     0.0000000     0.0000000     0.0000000     0.0000000     0.0000000");
    frm_modify_header(frm->line[87], newline);
    strcpy(newline, "ZEROADJ:     0.0000000     0.0000000     0.0000000     0.0000000");
    frm_modify_header(frm->line[88], newline);
    strcpy(newline, "XTRANS :     0.0000000     0.0000000     0.0000000");
    frm_modify_header(frm->line[89], newline);
    strcpy(newline, "HKL&XY :     0.0000000     0.0000000     0.0000000     0.0000000     0.0000000");
    frm_modify_header(frm->line[90], newline);
    strcpy(newline, "AXES2  :     0.0000000     0.0000000     0.0000000     0.0000000");
    frm_modify_header(frm->line[91], newline);
    strcpy(newline, "ENDING2:     0.0000000     0.0000000     0.0000000     0.0000000");
    frm_modify_header(frm->line[92], newline);
//    strcpy(newline, "       :");
//    frm_modify_header(frm->line[93], newline);
//    strcpy(newline, "       :");
//    frm_modify_header(frm->line[94], newline);
//    strcpy(newline, "       :");
//    frm_modify_header(frm->line[95], newline);
    frm->line[95][78] = 26;
    frm->line[95][79] = 4;

    // Now apply user specific header defaults
    frm_default_header_usersettings(frm);
    return 0;
}*/


/**+ -----------------------------------------------------------------------
FRM_SHRINK: Shrink frame image by factor of 2 in each direction.
*/
long frm_shrink(                /* (W) Error code: 0=none, 1=already 512 */
struct FRM_HEADER *frm,         /* (R/W) Frame header */
long frmdata[])                 /* (R/W) Image counts */
/**-*/
{
    char newline[81];
    float center[2], maxxy[2];
    long nrows, ncols, octmask[8], j, k, m, n;

    /* get large frame header parameters */
    nrows = atoi(frm->line[40]+8);
    ncols = atoi(frm->line[41]+8);
    center[0] = (float) atof(frm->line[54]+8);
    center[1] = (float) atof(frm->line[54]+8+14);
    maxxy[0] = (float) atof(frm->line[64]+8);
    maxxy[1] = (float) atof(frm->line[64]+8+14);
    octmask[0] = atoi(frm->line[74]+8);
    octmask[1] = atoi(frm->line[74]+8+12);
    octmask[2] = atoi(frm->line[74]+8+24);
    octmask[3] = atoi(frm->line[74]+8+36);
    octmask[4] = atoi(frm->line[74]+8+48);
    octmask[5] = atoi(frm->line[74]+8+60);
    octmask[6] = atoi(frm->line[75]+8);
    octmask[7] = atoi(frm->line[75]+8+12);

    /* check for minimum size of 512 x 512 image */
    if (ncols == 512) return 1;

    /* shrink data counts array */
    k = 0;
    m = 0;
    for (j=0; j<nrows; j=j+2) {
        for (n=0; n<ncols; n=n+2) {
            frmdata[k] = frmdata[m] + frmdata[m+1] + frmdata[m+ncols] + frmdata[m+ncols+1];
            k = k + 1;
            m = m + 2;
        }
        m = m + ncols;
    }

    /* update frame header items */
    ncols = ncols / 2;
    nrows = nrows / 2;
    center[0] = (center[0] + 0.5f) / 2.0f - 0.5f;
    center[1] = (center[1] + 0.5f) / 2.0f - 0.5f;
    maxxy[0] = (maxxy[0] + 0.5f) / 2.0f - 0.5f;
    maxxy[1] = (maxxy[1] + 0.5f) / 2.0f - 0.5f;
    for (n=0; n<8; n++)
        octmask[n] = octmask[n] / 2;

    /* place new values into frame header */
    sprintf(newline, "NROWS  :%10d", nrows);
    frm_modify_header(frm->line[40], newline);
    sprintf(newline, "NCOLS  :%10d", ncols);
    frm_modify_header(frm->line[41], newline);
    sprintf(newline, "CENTER :%14.7f%14.7f", center[0], center[1]);
    frm_modify_header(frm->line[54], newline);
    sprintf(newline, "MAXXY  :%14.7f%14.7f", maxxy[0], maxxy[1]);
    frm_modify_header(frm->line[64], newline);
    strcpy(newline, "BITMASK:$NULL");
    frm_modify_header(frm->line[73], newline);
    sprintf(newline, "OCTMASK:%12d%12d%12d%12d%12d%12d", octmask[0], octmask[1], octmask[2], octmask[3], octmask[4], octmask[5]);
    frm_modify_header(frm->line[74], newline);
    sprintf(newline, "OCTMASK:%12d%12d", octmask[6], octmask[7]);
    frm_modify_header(frm->line[75], newline);

    /* exit with error status */
    return 0;
}

void frm_convert_header_longs(
char *line72,           /* (R/W) 72-byte buffer */
long num,               /* (R) number of reals */
long bFreeFormat)       /* (R) TRUE for free format (Version 11) */
{
    char new72[73];
    long nval[6];
    long nread;

    // Read version 11 format
    nread = sscanf(line72, "%d %d %d %d %d %d", &nval[0], &nval[1], &nval[2],
        &nval[3], &nval[4], &nval[5]);

    // Write version 1-10 format
    switch (nread) {
    case 1:
        if (bFreeFormat) {
            sprintf(line72, "%-71ld", nval[0]);
            memset(line72+71, ' ', 1);
        }
        else {
            sprintf(line72, "%10d", nval[0]);
            memset(line72+10, ' ', 62);
        }
        break;
    case 2:
        if (bFreeFormat) {
            sprintf(line72, "%-35ld %-35ld", nval[0], nval[1]);\
            memset(line72+71, ' ', 1);
        }
        else {
            sprintf(line72, "%12d%12d", nval[0], nval[1]);\
            memset(line72+24, ' ', 48);
        }
        break;
    case 3:
        if (bFreeFormat) {
            sprintf(line72, "%-23ld %-23ld %-23ld", nval[0], nval[1], nval[2]);
            memset(line72+71, ' ', 1);
        }
        else {
            sprintf(line72, "%12d%12d%12d", nval[0], nval[1], nval[2]);
            memset(line72+36, ' ', 36);
        }
        break;
    case 4:
        if (bFreeFormat) {
            sprintf(line72, "%-17ld %-17ld %-17ld %-17ld", nval[0], nval[1], nval[2],
                nval[3]);
            memset(line72+71, ' ', 1);
        }
        else {
            sprintf(line72, "%12d%12d%12d%12d", nval[0], nval[1], nval[2],
                nval[3]);
            memset(line72+48, ' ', 24);
        }
        break;
    case 5:
        if (bFreeFormat) {
            sprintf(line72, "%-13ld %-13ld %-13ld %-13ld %-15ld", nval[0], nval[1], nval[2],
                nval[3], nval[4]);
            memset(line72+71, ' ', 1);
        }
        else {
            sprintf(line72, "%12d%12d%12d%12d%12d", nval[0], nval[1], nval[2],
                nval[3], nval[4]);
            memset(line72+60, ' ', 12);
        }
        break;
    case 6:
        if (bFreeFormat) {
            sprintf(line72, "%-11ld %-11ld %-11ld %-11ld %-11ld %-11ld", nval[0], nval[1], nval[2],
                nval[3], nval[4], nval[5]);
            memset(line72+71, ' ', 1);
        }
        else {
            sprintf(new72, "%12d%12d%12d%12d%12d%12d", nval[0], nval[1], nval[2],
                nval[3], nval[4], nval[5]);
            strncpy(line72, new72, 72);
        }
        break;
    }

    // Remove any NULLs (replace with spaces)
    //for (n=0; n<72; n++)
    //    if (line72[n] == '\0') line72[n] = ' ';

    return;
}

void frm_convert_header_reals(
char *line72,           /* (R/W) 72-byte buffer */
long num,               /* (R) number of reals */
long bFreeFormat)       /* (R) TRUE for free format (Version 11) */
{
    char new72[80];
    float real[5];
    long nread;

    // Read version 11 format
    nread = sscanf(line72, "%f%f%f%f%f", &real[0], &real[1], &real[2], &real[3],
        &real[4]);

    // Write version 1-10 format
    switch (nread) {
    case 1:
        if (bFreeFormat) {
            sprintf(line72, "%-71lf", real[0]);
            memset(line72+71, ' ', 1);
        }
        else {
            sprintf(line72, "%14.7f", real[0]);
            memset(line72+14, ' ', 58);
        }
        break;
    case 2:
        if (bFreeFormat) {
            sprintf(line72, "%-35lf %-35lf", real[0], real[1]);
            memset(line72+71, ' ', 1);
        }
        else {
            sprintf(line72, "%14.7f%14.7f", real[0], real[1]);
            memset(line72+28, ' ', 44);
        }
        break;
    case 3:
        if (bFreeFormat) {
            sprintf(line72, "%-23lf %-23lf %-23lf", real[0], real[1], real[2]);
            memset(line72+71, ' ', 1);
        }
        else {
            sprintf(line72, "%14.7f%14.7f%14.7f", real[0], real[1], real[2]);
            memset(line72+42, ' ', 30);
        }
        break;
    case 4:
        if (bFreeFormat) {
            sprintf(line72, "%-17lf %-17lf %-17lf %-17lf", real[0], real[1],
                real[2], real[3]);
            memset(line72+71, ' ', 1);
        }
        else {
            sprintf(line72, "%14.7f%14.7f%14.7f%14.7f", real[0], real[1],
                real[2], real[3]);
            memset(line72+56, ' ', 16);
        }
        break;
    case 5:
        if (bFreeFormat) {
            sprintf(line72, "%-13lf %-13lf %-13lf %-13lf %-15lf", real[0], real[1],
                real[2], real[3], real[4]);
            memset(line72+71, ' ', 1);
        }
        else {
            if (real[4] > 99999.9) {
                sprintf(new72, "%14.7f%14.7f%14.7f%14.7f %13f", real[0], real[1],
                    real[2], real[3], real[4]);
                strncpy(line72, new72, 72);
            }
            else
                sprintf(line72, "%14.7f%14.7f%14.7f%14.7f%14.7f", real[0], real[1],
                    real[2], real[3], real[4]);
            memset(line72+70, ' ', 2);
        }
        break;
    }

    // Remove any NULLs (replace with spaces)
    //for (n=0; n<72; n++)
    //    if (line72[n] == '\0') line72[n] = ' ';

    return;
}

/**+ -----------------------------------------------------------------------
FRM_CONVERT_HEADER:
      Converts VERSION 11 headers to VERSION 9 headers.
      Useful when converting FORMAT 100 to FORMAT 86 frames.
      Note: And reverse of above too.
*/
long frm_convert_header(        /* (W) Error code: 0=none, else=error */
struct FRM_HEADER *frm,         /* (R/W) FRM file header information */
long bFreeFormat)               /* (R) TRUE for free format (Version 11) */
/**-*/
{
    char newline[81];

    //frm$convert_header_longs(&frm->line[ 0][8], 1, bFreeFormat);  // FORMAT:
    if (bFreeFormat)
        sprintf(newline, "FORMAT :%-72ld", 100);
    else
        sprintf(newline, "FORMAT :%10d", 86);
    frm_modify_header(frm->line[0], newline);

    frm_convert_header_longs(&frm->line[ 1][8], 1, bFreeFormat);  // VERSION:
    //sprintf(newline, "VERSION:%10d", 9);
    //frm_modify_header(frm->line[1], newline);

    frm_convert_header_longs(&frm->line[ 2][8], 1, bFreeFormat);  // HDRBLKS:
    // TYPE:, SITE:, MODEL:, USER:, SAMPLE:, SETNAME: (no changes)
    frm_convert_header_longs(&frm->line[ 9][8], 1, bFreeFormat);  // RUN:
    frm_convert_header_longs(&frm->line[10][8], 1, bFreeFormat);  // SAMPNUM:
    // TITLE: (no changes)
    frm_convert_header_longs(&frm->line[19][8], 2, bFreeFormat);  // NCOUNTS: (ignore others)
    frm_convert_header_longs(&frm->line[20][8], 3, bFreeFormat);  // NOVERFL: (will change)
    frm_convert_header_longs(&frm->line[21][8], 1, bFreeFormat);  // MINIMUM:
    frm_convert_header_longs(&frm->line[22][8], 1, bFreeFormat);  // MAXIMUM:
    frm_convert_header_longs(&frm->line[23][8], 1, bFreeFormat);  // NONTIME:
    frm_convert_header_longs(&frm->line[24][8], 1, bFreeFormat);  // NLATE:
    // FILENAM:, CREATED: (no changes)
    frm_convert_header_reals(&frm->line[27][8], 1, bFreeFormat);  // CUMALAT:
    frm_convert_header_reals(&frm->line[28][8], 1, bFreeFormat);  // ELAPSDR:
    frm_convert_header_reals(&frm->line[29][8], 1, bFreeFormat);  // ELAPSDA:
    frm_convert_header_longs(&frm->line[30][8], 1, bFreeFormat);  // OSCILLA:
    frm_convert_header_longs(&frm->line[31][8], 1, bFreeFormat);  // NSTEPS:
    frm_convert_header_reals(&frm->line[32][8], 1, bFreeFormat);  // RANGE:
    frm_convert_header_reals(&frm->line[33][8], 1, bFreeFormat);  // START:
    frm_convert_header_reals(&frm->line[34][8], 1, bFreeFormat);  // INCREME:
    frm_convert_header_longs(&frm->line[35][8], 1, bFreeFormat);  // NUMBER:
    frm_convert_header_longs(&frm->line[36][8], 1, bFreeFormat);  // NFRAMES:
    frm_convert_header_reals(&frm->line[37][8], 4, bFreeFormat);  // ANGLES:
    frm_convert_header_longs(&frm->line[38][8], 3, bFreeFormat);  // NOVER64K: (will change)
    frm_convert_header_longs(&frm->line[39][8], 2, bFreeFormat);  // NPIXELB: (will change)
    frm_convert_header_longs(&frm->line[40][8], 1, bFreeFormat);  // NROWS:
    frm_convert_header_longs(&frm->line[41][8], 1, bFreeFormat);  // NCOLS:
    frm_convert_header_longs(&frm->line[42][8], 1, bFreeFormat);  // WORDORD:
    frm_convert_header_longs(&frm->line[43][8], 1, bFreeFormat);  // LONGORD:
    // TARGET: (no changes)
    frm_convert_header_reals(&frm->line[45][8], 1, bFreeFormat);  // SOURCEK:
    frm_convert_header_reals(&frm->line[46][8], 1, bFreeFormat);  // SOURCEM:
    // FILTER: (no changes)
    frm_convert_header_reals(&frm->line[48][8], 5, bFreeFormat);  // CELL:
    frm_convert_header_reals(&frm->line[49][8], 1, bFreeFormat);  // CELL:
    frm_convert_header_reals(&frm->line[50][8], 5, bFreeFormat);  // MATRIX:
    frm_convert_header_reals(&frm->line[51][8], 4, bFreeFormat);  // MATRIX:
    frm_convert_header_longs(&frm->line[52][8], 3, bFreeFormat);  // LOWTEMP:
    frm_convert_header_reals(&frm->line[53][8], 3, bFreeFormat);  // ZOOM:
    frm_convert_header_reals(&frm->line[54][8], 4, bFreeFormat);  // CENTER:
    frm_convert_header_reals(&frm->line[55][8], 1, bFreeFormat);  // DISTANC:
    // TRAILER: (unused)
    // COMPRES: (no changes)
    frm_convert_header_reals(&frm->line[58][8], 2, bFreeFormat);  // LINEAR:
    frm_convert_header_reals(&frm->line[59][8], 2, bFreeFormat);  // PHD:
    frm_convert_header_reals(&frm->line[60][8], 1, bFreeFormat);  // PREAMP:
    // CORRECT:, WARPFIL: (no changes)
        {
            long bIsBlank;
            char szLine[80];
            char *ptr;

            memset(szLine, 0, sizeof(szLine));
            strncpy(szLine, &frm->line[61][8], 72);
            strupr(szLine);
            ptr = strstr(szLine, "\\LINEAR");
            bIsBlank = (szLine[0] == '\0') || 
                (strnicmp(szLine, "                ", 16) == 0);
            if ((ptr != NULL) || bIsBlank) {
                sprintf(newline, "CORRECT:LINEAR");
                frm_modify_header(frm->line[61], newline);
            }

            memset(szLine, 0, sizeof(szLine));
            strncpy(szLine, &frm->line[62][8], 72);
            strupr(szLine);
            ptr = strstr(szLine, "\\LINEAR");
            bIsBlank = (szLine[0] == '\0') || 
                (strnicmp(szLine, "                ", 16) == 0);
            if ((ptr != NULL) || bIsBlank) {
                sprintf(newline, "WARPFIL:LINEAR");
                frm_modify_header(frm->line[62], newline);
            }
        }

    frm_convert_header_reals(&frm->line[63][8], 4, bFreeFormat);  // WAVELEN:
    if (frm->line[64][7] == ':')
        frm_convert_header_reals(&frm->line[64][8], 2, bFreeFormat);  // MAXXY:
    if (frm->line[65][7] == ':')
        frm_convert_header_longs(&frm->line[65][8], 1, bFreeFormat);  // AXIS:
    if (frm->line[66][7] == ':')
        frm_convert_header_reals(&frm->line[66][8], 4, bFreeFormat);  // ENDING:
    if (frm->line[67][7] == ':')
        frm_convert_header_reals(&frm->line[67][8], 5, bFreeFormat);  // DETPAR:
    if (frm->line[68][7] == ':')
        frm_convert_header_reals(&frm->line[68][8], 1, bFreeFormat);  // DETPAR:
    // LUT: (no changes)
    if (frm->line[70][7] == ':')
        frm_convert_header_reals(&frm->line[70][8], 2, bFreeFormat);  // DISPLIM:
    // PROGRAM: (no changes)
    if (frm->line[72][7] == ':')
        frm_convert_header_longs(&frm->line[72][8], 1, bFreeFormat);  // ROTATE:
    // BITMASK: (no changes)
    if (frm->line[74][7] == ':')
        frm_convert_header_longs(&frm->line[74][8], 6, bFreeFormat);  // OCTMASK:
    if (frm->line[75][7] == ':')
        frm_convert_header_longs(&frm->line[75][8], 2, bFreeFormat);  // OCTMASK:
    if (frm->line[76][7] == ':')
        frm_convert_header_reals(&frm->line[76][8], 5, bFreeFormat);  // ESDCELL:
    if (frm->line[77][7] == ':')
        frm_convert_header_reals(&frm->line[77][8], 1, bFreeFormat);  // ESDCELL:
    // DETTYPE: (no changes, read in free-format)
    if (frm->line[79][7] == ':')
        frm_convert_header_longs(&frm->line[79][8], 5, bFreeFormat);  // NEXP:
    if (frm->line[80][7] == ':')
        frm_convert_header_reals(&frm->line[80][8], 5, bFreeFormat);  // CCDPARM:
    // CHEM:, MORPH:, CCOLOR:, CSIZE:, DNSMET:, DARK: (no changes)
    if (frm->line[87][7] == ':')
        frm_convert_header_reals(&frm->line[87][8], 5, bFreeFormat);  // AUTORNG:
    if (frm->line[88][7] == ':')
        frm_convert_header_reals(&frm->line[88][8], 4, bFreeFormat);  // ZEROADJ:
    if (frm->line[89][7] == ':')
        frm_convert_header_reals(&frm->line[89][8], 3, bFreeFormat);  // XTRANS:
    if (frm->line[90][7] == ':')
        frm_convert_header_reals(&frm->line[90][8], 5, bFreeFormat);  // HKL&XY:
    if (frm->line[91][7] == ':')
        frm_convert_header_reals(&frm->line[91][8], 4, bFreeFormat);  // AXES2:
    if (frm->line[92][7] == ':')
        frm_convert_header_reals(&frm->line[92][8], 4, bFreeFormat);  // ENDING2:
    if (frm->line[93][7] == ':')
        frm_convert_header_reals(&frm->line[93][8], 2, bFreeFormat);  // FILTER2:

    return 0;
}

/**+ -----------------------------------------------------------------------
FRM_UPDATE_HEADER:
      Converts older VERSION headers to requested VERSION headers.
      Note: Does not downgrade headers to older versions.
*/
long frm_update_header(         /* (W) Error code: 0=none, else=error */
struct FRM_HEADER *frm,         /* (R/W) FRM file header information */
long nReqVersion)               /* (R) Requested header version level */
/**-*/
{
    char newline[81];
    long nOldVersion = atoi(&frm->line[1][8]);

    // Check if already done
    if (nReqVersion <= nOldVersion)
        return 0;

    // Update to Version 11 header add FILTER2 line
    if ((nOldVersion < 11) && (nReqVersion > 10)) {
        sprintf(newline, "FILTER2:%-35lf %-35lf", -1., -1.);
        frm_modify_header(frm->line[93], newline);
    }

    return 0;
}

/**+ -----------------------------------------------------------------------
FRM_TRIM: Trim frame image to Bruker frame dimensions in each direction.
*/
long frm_trim(                  /* (W) Error code: 0=none, 1=already 512 */
struct FRM_HEADER *frm,         /* (R/W) Frame header */
long frmdata[])                 /* (R/W) Image counts */
/**-*/
{
    char newline[81];
    float center[2], maxxy[2];
    long nrows, ncols, j, k, m; //, octmask[8]
    long nsize, l, r, t, b;

    /* get large frame header parameters */
    nrows = atoi(frm->line[40]+8);
    ncols = atoi(frm->line[41]+8);
    center[0] = (float) atof(frm->line[54]+8);
    center[1] = (float) atof(frm->line[54]+8+14);
    maxxy[0] = (float) atof(frm->line[64]+8);
    maxxy[1] = (float) atof(frm->line[64]+8+14);
    //octmask[0] = atoi(frm->line[74]+8);
    //octmask[1] = atoi(frm->line[74]+8+12);
    //octmask[2] = atoi(frm->line[74]+8+24);
    //octmask[3] = atoi(frm->line[74]+8+36);
    //octmask[4] = atoi(frm->line[74]+8+48);
    //octmask[5] = atoi(frm->line[74]+8+60);
    //octmask[6] = atoi(frm->line[75]+8);
    //octmask[7] = atoi(frm->line[75]+8+12);

    /* check for minimum size of 512 x 512 image */
    if ((nrows == 512) && (ncols == 512)) return 1;
    nsize = min(nrows, ncols);
    nsize = (nsize / 512) * 512;

    /* trim parameters */
    l = (ncols - nsize) / 2;
    r = ncols - nsize - l;
    t = (nrows - nsize) / 2;
    b = nrows - nsize - t;
    //if (dbglevel() > 1) {
        printf("Trimming image: Left: %d  Right: %d  Top: %d  Bottom: %d\n", l, r, t, b);
    //}

    /* trim data counts array (in-place) */
    for (j=0; j<nsize; j++) {
        k = j * nsize;
        m = (j+t) * ncols + l;
        memmove(&frmdata[k], &frmdata[m], nsize*sizeof(long));
    }
    if (dbglevel() > 1) {
        printf("Trimming image: Done\n");
    }

    /* update frame header items */
    ncols = nsize;
    nrows = nsize;
    center[0] = center[0] - l;
    center[1] = center[1] - b;
    maxxy[0] = maxxy[0] - l;
    maxxy[1] = maxxy[1] - b;
    //for (n=0; n<8; n++)
    //    octmask[n] = octmask[n] / 2;

    /* place new values into frame header */
    sprintf(newline, "NROWS  :%10d", nrows);
    frm_modify_header(frm->line[40], newline);
    sprintf(newline, "NCOLS  :%10d", ncols);
    frm_modify_header(frm->line[41], newline);
    sprintf(newline, "CENTER :%14.7f%14.7f", center[0], center[1]);
    frm_modify_header(frm->line[54], newline);
    sprintf(newline, "MAXXY  :%14.7f%14.7f", maxxy[0], maxxy[1]);
    frm_modify_header(frm->line[64], newline);
    strcpy(newline, "BITMASK:$NULL");
    frm_modify_header(frm->line[73], newline);
    //sprintf(newline, "OCTMASK:%12d%12d%12d%12d%12d%12d", octmask[0], octmask[1], octmask[2], octmask[3], octmask[4], octmask[5]);
    //frm_modify_header(frm->line[74], newline);
    //sprintf(newline, "OCTMASK:%12d%12d", octmask[6], octmask[7]);
    //frm_modify_header(frm->line[75], newline);

    /* exit with error status */
    return 0;
}

/**+ -----------------------------------------------------------------------
FRM_TRUNCATE: Truncate image to smaller size (upper left quadrant becomes
    full image). Useful, when 512 spatial used on 1024 image.
*/
long frm_truncate(              /* (W) Error code: 0=none, 1=already 512 */
struct FRM_HEADER *frm,         /* (R/W) Frame header */
long frmdata[])                 /* (R/W) Image counts */
/**-*/
{
    char newline[81];
    float center[2], maxxy[2];
    long nrows, ncols, j, k, m, octmask[8];
    long nsize, l, r, t, b, n;

    /* get large frame header parameters */
    nrows = atoi(frm->line[40]+8);
    ncols = atoi(frm->line[41]+8);
    center[0] = (float) atof(frm->line[54]+8);
    center[1] = (float) atof(frm->line[54]+8+14);
    maxxy[0] = (float) atof(frm->line[64]+8);
    maxxy[1] = (float) atof(frm->line[64]+8+14);
    octmask[0] = atoi(frm->line[74]+8);
    octmask[1] = atoi(frm->line[74]+8+12);
    octmask[2] = atoi(frm->line[74]+8+24);
    octmask[3] = atoi(frm->line[74]+8+36);
    octmask[4] = atoi(frm->line[74]+8+48);
    octmask[5] = atoi(frm->line[74]+8+60);
    octmask[6] = atoi(frm->line[75]+8);
    octmask[7] = atoi(frm->line[75]+8+12);

    /* check for minimum size of 512 x 512 image */
    if ((nrows == 512) && (ncols == 512)) return 1;
    nsize = min(nrows, ncols);
    nsize = (nsize / 512) * 512;
    nsize = nsize / 2;

    /* trim parameters */
    l = 0;
    r = nsize;
    t = 0;
    b = nsize;
    //if (dbglevel() > 1) {
        printf("Truncating image: Left: %d  Right: %d  Top: %d  Bottom: %d\n", l, r, t, b);
    //}

    /* trim data counts array (in-place) */
    for (j=0; j<nsize; j++) {
        k = j * nsize;
        m = (j+t) * ncols + l;
        memmove(&frmdata[k], &frmdata[m], nsize*sizeof(long));
    }
    if (dbglevel() > 1) {
        printf("Truncating image: Done\n");
    }

    /* update frame header items */
    ncols = ncols / 2;
    nrows = nrows / 2;
    center[0] = center[0] / 2.0f;
    center[1] = center[1] / 2.0f;
    maxxy[0] = maxxy[0] / 2.0f;
    maxxy[1] = maxxy[1] / 2.0f;
    for (n=0; n<8; n++)
        octmask[n] = octmask[n] / 2;

    /* place new values into frame header */
    sprintf(newline, "NROWS  :%10d", nrows);
    frm_modify_header(frm->line[40], newline);
    sprintf(newline, "NCOLS  :%10d", ncols);
    frm_modify_header(frm->line[41], newline);
    sprintf(newline, "CENTER :%14.7f%14.7f", center[0], center[1]);
    frm_modify_header(frm->line[54], newline);
    sprintf(newline, "MAXXY  :%14.7f%14.7f", maxxy[0], maxxy[1]);
    frm_modify_header(frm->line[64], newline);
    strcpy(newline, "BITMASK:$NULL");
    frm_modify_header(frm->line[73], newline);
    sprintf(newline, "OCTMASK:%12d%12d%12d%12d%12d%12d", octmask[0], octmask[1], octmask[2], octmask[3], octmask[4], octmask[5]);
    frm_modify_header(frm->line[74], newline);
    sprintf(newline, "OCTMASK:%12d%12d", octmask[6], octmask[7]);
    frm_modify_header(frm->line[75], newline);

    /* exit with error status */
    return 0;
}

/**+ -----------------------------------------------------------------------
FRM_REORIENTATE: Reorientate image and update header items.
*/
long frm_reorientate(long nOrientation,struct FRM_HEADER *frm,long *frmdata)
{
    char newline[81];
    float center[2], maxxy[2];
    float oldcenter[2], oldmaxxy[2];
    long nrows, ncols, octmask[8], nsize;
    long k, j, x, y, *pTmp;

    // Check validity of input parameters
    switch (nOrientation) {
    case 1:     /* +x, +y */
        printf("FRM2FRM-I-Already in requested orientation.\n");
        return 0;
    case 2:     /* +x, -y */
    case 3:     /* -x, +y */
    case 4:     /* -x, -y */
    case 5:     /* +y, +x */
    case 6:     /* +y, -x */
    case 7:     /* -y, +x */
    case 8:     /* -y, -x */
        break;
    default:    /* invalid */
        printf("FRM2FRM-F-Bad orientation request.\n");
        return -1;
    }

    /* get frame header parameters */
    nrows = atoi(frm->line[40]+8);
    ncols = atoi(frm->line[41]+8);
    oldcenter[0] = (float) atof(frm->line[54]+8);
    oldcenter[1] = (float) atof(frm->line[54]+8+14);
    oldmaxxy[0] = (float) atof(frm->line[64]+8);
    oldmaxxy[1] = (float) atof(frm->line[64]+8+14);
    octmask[0] = atoi(frm->line[74]+8);
    octmask[1] = atoi(frm->line[74]+8+12);
    octmask[2] = atoi(frm->line[74]+8+24);
    octmask[3] = atoi(frm->line[74]+8+36);
    octmask[4] = atoi(frm->line[74]+8+48);
    octmask[5] = atoi(frm->line[74]+8+60);
    octmask[6] = atoi(frm->line[75]+8);
    octmask[7] = atoi(frm->line[75]+8+12);

    /* Make a copy of frame image data */
    nsize = nrows * ncols * sizeof(unsigned long);
    pTmp = (long *) malloc(nsize);
    if (pTmp == NULL) return -1;
    memcpy(pTmp, frmdata, nsize);

    // Re-orientate image
    switch (nOrientation) {
    case 2:     /* +x, -y */
        k = 0; j= 0;
        for (y=0; y<nrows; y++) {
            j = (nrows-1-y) * ncols;
            for (x=0; x<ncols; x++) {
                frmdata[k] = pTmp[j];
                k++; j++;
            }
        }
        center[0] = oldcenter[0];
        center[1] = (nrows-1) - oldcenter[1];
        maxxy[0] = oldmaxxy[0];
        maxxy[1] = (nrows-1) - oldmaxxy[1];
        //octmask is n.y.i.
        break;

    case 3:     /* -x, +y */
        k = 0; j= 0;
        for (y=0; y<nrows; y++) {
            for (x=0; x<ncols; x++) {
                j = y*ncols + (ncols-1)-x;
                frmdata[k] = pTmp[j];
                k++;
            }
        }
        center[0] = (ncols-1) - oldcenter[0];
        center[1] = oldcenter[1];
        maxxy[0] = (ncols-1) - oldmaxxy[0];
        maxxy[1] = oldmaxxy[1];
        //octmask is n.y.i.
        break;

    case 4:     /* -x, -y */
        k = 0; j= 0;
        for (y=0; y<nrows; y++) {
            for (x=0; x<ncols; x++) {
                j = (nrows-1-y) * ncols + (ncols-1)-x;
                frmdata[k] = pTmp[j];
                k++;
            }
        }
        center[0] = (ncols-1) - oldcenter[0];
        center[1] = (nrows-1) - oldcenter[1];
        maxxy[0] = (ncols-1) - oldmaxxy[0];
        maxxy[1] = (nrows-1) - oldmaxxy[1];
        //octmask is n.y.i.
        break;

    case 5:     /* +y, +x */
        k = 0; j= 0;
        for (y=0; y<nrows; y++) {
            for (x=0; x<ncols; x++) {
                j = x * ncols + y;
                frmdata[k] = pTmp[j];
                k++;
            }
        }
        center[0] = (nrows-1) - oldcenter[1];
        center[1] = (ncols-1) - oldcenter[0];
        maxxy[0] = (nrows-1) - oldmaxxy[1];
        maxxy[1] = (ncols-1) - oldmaxxy[0];
        //octmask is n.y.i.
        break;

    case 6:     /* +y, -x */
        k = 0; j= 0;
        for (y=0; y<nrows; y++) {
            for (x=0; x<ncols; x++) {
                j = (nrows-1-x) * ncols + y;
                frmdata[k] = pTmp[j];
                k++;
            }
        }
        center[0] = oldcenter[1];
        center[1] = (nrows-1) - oldcenter[0];
        maxxy[0] = oldmaxxy[1];
        maxxy[1] = (nrows-1) - oldmaxxy[0];
        //octmask is n.y.i.
        break;

    case 7:     /* -y, +x */
        k = 0; j= 0;
        for (y=0; y<nrows; y++) {
            for (x=0; x<ncols; x++) {
                j = x * ncols + (ncols-1 - y);
                frmdata[k] = pTmp[j];
                k++;
            }
        }
        center[0] = (ncols-1) - oldcenter[1];
        center[1] = oldcenter[0];
        maxxy[0] = (ncols-1) - oldmaxxy[1];
        maxxy[1] = oldmaxxy[0];
        //octmask is n.y.i.
        break;

    case 8:     /* -y, -x */
        k = 0; j= 0;
        for (y=0; y<nrows; y++) {
            for (x=0; x<ncols; x++) {
                j = (nrows-1 - x) * ncols + (ncols-1 - y);
                frmdata[k] = pTmp[j];
                k++;
            }
        }
        center[0] = oldcenter[1];
        center[1] = oldcenter[0];
        maxxy[0] = oldmaxxy[1];
        maxxy[1] = oldmaxxy[0];
        //octmask is n.y.i.
        break;

    }   // end switch (nOrientation)

    /* Clean-up */
    free (pTmp);

    /* place new values into frame header */
    sprintf(newline, "CENTER :%14.7f%14.7f", center[0], center[1]);
    frm_modify_header(frm->line[54], newline);
    sprintf(newline, "MAXXY  :%14.7f%14.7f", maxxy[0], maxxy[1]);
    frm_modify_header(frm->line[64], newline);
    strcpy(newline, "BITMASK:$NULL");
    frm_modify_header(frm->line[73], newline);
    sprintf(newline, "OCTMASK:%12d%12d%12d%12d%12d%12d", octmask[0], octmask[1], octmask[2], octmask[3], octmask[4], octmask[5]);
    frm_modify_header(frm->line[74], newline);
    sprintf(newline, "OCTMASK:%12d%12d", octmask[6], octmask[7]);
    frm_modify_header(frm->line[75], newline);

    /* exit with error status */
    return 0;
}
