// DiffractionImage.h
// maintained by G.Winter
// 15th December 2003
// 
// A replacement for the old DiffractionImage library - with one 
// of the same name and a very similar API. Take care though - this one 
// will be different!

// switch on the GSL stuff
#define INCLUDE_GSL
#ifndef DIFFRACTIONIMAGE_HEADER
#define DIFFRACTIONIMAGE_HEADER

#ifdef _MSC_VER
#define MYREAD O_RDONLY|O_BINARY
#define MYWRITE O_WRONLY|O_BINARY
#define MYAPPEND O_APPEND|O_BINARY
#else
#define MYREAD O_RDONLY
#define MYWRITE O_WRONLY
#define MYAPPEND O_APPEND
#endif

#ifdef TEST_CODE
#define NDEBUG
#include <assert.h>
#endif

#ifndef TEST_CODE
#define GET_FROM_HEADER(key, format, value)	\
    sscanf(headerMap[key].c_str(), format, &value);
#define ASSERT_IN_HEADER(key)		    \
    if (headerMap.find(key) == headerMap.end()) \
      {								  \
	throw(DiffractionImageException(key " not found in header"));	\
      }
#else
#define ASSERT_IN_HEADER(key)		    \
    if (headerMap.find(key) == headerMap.end()) \
      {								  \
	throw(DiffractionImageException(key " not found in header"));	\
      } else { cout << "found key \"" << key << "\"" << endl; }
#define GET_FROM_HEADER(key, format, value)	\
    sscanf(headerMap[key].c_str(), format, &value);	\
    cout << key << "=" << value << endl;
#endif


#include <iostream>
#include <fstream>
#include <cstring>
#include <vector>
#include <map>
#include <list>
#include <new>
#include <cmath>
#include <cstdlib>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctime>
#if defined _MSC_VER
#include <io.h>
#else
#include <unistd.h>
#endif

// include helpful things from the standard namespace
// like cout and cerr.

using namespace std;

#ifdef INCLUDE_PYTHON
#include "Python.h"
#endif

#ifdef SWIG
int atPos(int pos,int* array);
int atPos(int pos,unsigned int* array);
unsigned short atPos(int pos,unsigned short* array);
float atPos(int pos,float* array);
#endif

namespace DI 
	{
    class DiffractionImage 
		{
		public:
		// public constructors and destructors
		DiffractionImage(void);             // done
		DiffractionImage(char* filename);  // done
		~DiffractionImage(void);            // done
	
		// constructors and destructors, volume II - this
		// time they are real constructors etc, which the above
		// will call - this enables the code for Python objects
		// share the same constructors etc without duplicating 
		// code
		// the destructor will re-initialise the structure, so this
		// could be used for resetting
		void constructor(void);            // done
		void constructor(char* filename); // done
		void destructor(void);             // done

		// next the get & set methods - these should be generated 
		// and not hand coded, but I can't be bothered.
		// note well - I am trying with this to not be fussy - 
		// so I will avoid unsigned, const and references where
		// possible
		// these should be defined via macro to save typing!
		// never mind.. setter methods should also be available
		// to allow external functions to be able to read and assign 
		// images.
		int getWidth(void);
		int getHeight(void);
		float getBeamX(void);
		float getBeamY(void);
		float getWavelength(void);
		float getDistance(void);
		float getTwoTheta(void);
		float getOscStart(void);
		float getOscEnd(void);
		float getPixelX(void);
		float getPixelY(void);
		float getExposureTime(void);
		float* getR();
		float* getS();
		float* getP();
		bool getWarning();
        const char* getMessage();
		const char* getDate();
		const char* getSerialNo();
        const char* getFormat();
		const char* getManufacturer();
		const char* getFilename();
		const char* getOscAxis();
        // get a pointer to the image - and use with great care!
		// the image will be stored as an unsigned short, as this 
		// seems to be the common storage type
		unsigned int* getImage(void);
		
		int getPixelAt(int x,int y);

		//returns the filename of the temporary ppm image
		char* getImage(double rate,int upleftX,int upleftY,int boxW,int boxH);

		// setter methods for the same data
		void setBeamX(float);
		void setBeamY(float);
		void setWavelength(float);
		void setDistance(float);
		void setTwoTheta(float);
		void setOscStart(float);
		void setOscEnd(float);
		void setOscAxis(char*);
		void setPixelX(float);
		void setPixelY(float);
		void setExposureTime(float);
		void setImage(unsigned int* copy);
		void setWidth(int);
		void setHeight(int);
		bool isOriginalImg();
	
		// some routines for comparing images - min, max, diff, sum
		// will alter THIS image
		void min(DiffractionImage* other);
		void max(DiffractionImage* other);
		void diff(DiffractionImage* other);
		void sum(DiffractionImage* other);
		// next - reading and writing methods - these depend in some values
		// defined in the header struture of the images
		// redefine this to return a string bearing the image type
		// will return one of "adsc", "marccd" and "mar" at the moment
        void load(char* fileName);
        void loadHeader(char* fileName);
		// this will assume that the image and header are called
		// imageName + .jp2, imageName.hdr respectively
		void loadCompressed(string imageName);
		void loadCompressed(string headerFileName,string imageFileName);
		// this will JUST read the jp2 image - DO NOT USE ALONE
		void loadCompressedImage(string imageName);

        // this one is ambitious! I want to allow header editing - but 
		// this should not promise to retain all of the info in the header.
		void regenerateHeader(void);
        
		// calculations you can do with a DiffractionImage
        int* histogram(void);
		float* radial(void);
		float* radial(int bins);
		float* asymmetry(void);
		int corner(void);
		double gain(void);
		int gainX(void);
		int gainY(void);
		float resolRingRadius(float resolution);
		float resolOnRadius(float radius);
		char* autoMask(int beamX, int beamY);
        
        // methods to write the image out as a jpeg
		// these are void and so will raise an exception
		// in the event of an error
		// by default this will produce a JPEG image with quality
		// 85 (good enough) and size somewhere around 1k x 1k.
		// these are done=
		void jpeg(string filename);
		void jpeg(string filename,int quality);
		void jpeg(string filename,int quality,int zoom);
		
		// if optimise == true then do floating point, 
		// else just shift bits
		void jpeg(string filename,int quality,int zoom,bool optmise);
        // array subscription operators
		// by design these MUST return references
#ifndef SWIG
        unsigned int & operator[] (int offset);
	
		// copy constructors & etc - likewise require references
		DiffractionImage & operator= (DiffractionImage & otherImage);
#endif

    private:
		// this is the real data which represents my image structure
		// I have yet to meet a header > 4k
		int newHeaderSize();
        //these function should not be accessible to outside world for
        //security issues filename does not need to be visible
        //and surely outside world should not be able to change them in the process
        void setFormat(string);
		void setFilename(string);
        //these are only used internally and therefore do not need to be public
        void load();
        void loadHeader();
        // specific methods to read images
		void allocateImage(void); // allocate the image according to the information in the image header
		void loadRawImage(string filename,int offset);
		void loadADSC();
		void loadCBF();
		void loadRaxis();
		void loadRAXISDtrek();
		void loadMarCCD(string filename);
		void loadMAR();
		void loadSaturn();
		void loadBruker();
		void loadOxford();
		void loadTIFF();
		void loadADSCHeader();
		void loadCBFHeader();
		void loadRaxisHeader();
		void loadRAXISDtrekHeader();
		void loadMarCCDHeader(string filename);
		void loadMARHeader();
		void loadSaturnHeader();
		void loadBrukerHeader();
		void loadOxfordHeader();
		void loadTIFFHeader();
		
		//These function are all related to the automatic mask generation.
		float* backstopDiscFit(int startX, int startY);
		int* simpleArmFit(int* from, int* to,float angle,float bckstopRad);
		int* complexArmFit(int* from, int* to,float angle,float bckstopRad);
		int* lowestPixelAround(int fromX, int fromY,int vecX, int vecY);
		float averageDisc(float cx, float cy, float r);
		float varianceDisc(float cx, float cy, float r);
		float averageQuad(int* vertices);
		float varianceQuad(int* vertices);
		float rayAverage(int fromX,int fromY,double vecX,double vecY,int steps);
        float rayVariance(int fromX,int fromY,double vecX,double vecY,int steps);
		int ray(int fromX, int fromY, double vecX, double vecY,double thresh);
		float arcAverage(float rad,float startAng,float endAng);
        float arcVariance(float rad,float startAng,float endAng);
		float arcRay(float fromAng, float radius, bool plus, double thresh);
		bool rayOnTileJoint(int fromX, int fromY,double vecX, double vecY);
	   	bool lineIsBlank(int start,bool horizontal);
		bool circleIsBlank(int centX,int centY,float radius);
		int* cleanEdgeVector(int edgeX,int edgeY);
		bool isInBadZone(int x, int y);
		int* correctCoord(int x, int y);
		
		unsigned char originalHeader[4096];
		int originalHeaderLength;
		bool bigEndian; // this will be true if the original image was big endian

		// the basic image size
		int width;
		int height;
		int gaintopX;
		int gaintopY;
		float beamX;
		float beamY;
		float distance;
		float wavelength;
		float twoTheta;
		float oscStart;
		float oscEnd;
		float pixelX;
		float pixelY;
		map <string, string> headerMap;
		vector <string> headerKeys;
		list<int> maxXs;
		list<int> maxYs;
		string date;
		string oscAxis;
		string serialNo;
		float exposureTime; // in seconds
		unsigned int* image;
		float rapidAxis[3];
		float slowAxis[3];
		float pixelOrigin[3];
		float imgMax;
		bool realImage;

		// a way of communicating between the image loading
		// expertise and the outside world!
		bool warning;
		string message;
		string format;
		string manufact;
		string filename;
    };

    class DiffractionImageException 
		{
		public:
			DiffractionImageException(void);
			DiffractionImageException(string newMessage);
			~DiffractionImageException(void);
			string getMessage(void);
		private:
			string message;
		};
    
    // guessFormat will return either "adsc", "marccd", "cbf, "raxis, "unknown"
    string guessFormat(string file);
	
	// other functions
    void enlargeImage(unsigned int * original,int width,int height,int zoom,unsigned int * copy); // this must be allocated
    void reduceImage(unsigned int * original,int width,int height,int zoom,unsigned int * copy); // this must be allocated

    // this one will also do the downsampling to eight bit depth
    // if optimise then perform some optimisation i.e. looking at the
    // spread of values else just / 256
    void downSample(unsigned int * original,int width,int height,bool optimise,unsigned char * copy);
    void invertImage(unsigned char * image,int width,int height);

    // this will include the downsampling and the image inversion
    void writeJpegImage(unsigned int * image,int width,int height,int quality,bool optimise,const char * filename);

    // some utility routines - these are defined in 
    // DiffractionImageUtility.cpp as functions
    bool machineBigEndian(void);
    void byteSwap(unsigned short * blob);
    void byteSwap(unsigned int * blob);
	
	int fitCircle(float x0,float y0,float x1,float y1,float x2,float y2,
	              float & x,float & y,float & r);
				 
	bool isInsideQuad(int* quad,int pointX,int pointY);
	bool isInsideTriangle(int* triangle,int pointX,int pointY);
	};
#endif
