// DiffractionImageCalculations.cpp
// maintained by F.Remacle
// 4th June 2007
// 
// The implementation of various functions that take part of the automaic masking
// functions from the calculations
//
// $Id: DiffractionImageAutomask.cpp,v 1.12 2008/07/22 15:01:27 FR45 Exp $

#include "DiffractionImage.h"

namespace DI
	{
	float* DiffractionImage::backstopDiscFit(int startX, int startY)
		{
		int ind,nsteps,fitX,fitY,ncircles;
		int sum,sumsq,npix,maxX,maxY;
		unsigned int maxVal;
		float average,variance,oldvariance,fitR,endX,endY,endR;
		double* distances;
		double magic;
		float* returnval;
		int* coords;
		bool trygrow;
			
		returnval=new float[3];
		coords=new int[6];
		distances=new double[3];
		maxXs.clear();
		maxYs.clear();
		
		ncircles=0;
		fitX=0;
		fitY=0;
		fitR=0;
		if(isOriginalImg())
			magic=0.333;
		else
			magic=0.555;
		
		//First we are going to prescan the area around the disc zone to rule out
		//high peak in this zone by keeping a list of distinct risk locations.
		//This is to try to avoid taking artifacts and beam leakage into account
		//when computing mean and std-dev in the fitting.
		sum=0;
		sumsq=0;
		npix=0;
		maxVal=0;
		maxX=-1;
		maxY=-1;
		
		//Firstly we loop over to calulate the mean on the area we are going to scan.
		//we keep the maximum location by the same time.
		for(int j=startY-(int)floor(0.025*height);j<startY+(int)ceil(0.025*height);j++)
			{
			for(int i=startX-(int)floor(0.025*width);i<startX+(int)ceil(0.025*width);i++)
				{
				if(image[j*width+i] > 0)
					{
					sum+=image[j*width+i];
					npix++;
					}
				if(image[j*width+i] > maxVal)
					{
					
					maxVal=image[j*width+i];
					maxX=i;
					maxY=j;
					}
				}
			}
		
		average=sum/(npix*1.0f);
		if(maxVal>5*average)
			{
			maxXs.push_back(maxX);
			maxYs.push_back(maxY);
			printf("Ignoring 20x20 box around (%d,%d)\n",maxX,maxY);
			}
		//Now we need to rescan the area to look for peaks that stands out from 
		//the average value.
		for (int h=0;h<2;h++)
			{
			maxVal=0;
			maxX=-1;
			maxY=-1;
			for(int i=startX-(int)floor(0.025*width);i<startX+(int)ceil(0.025*width);i++)
				{
				for(int j=startY-(int)floor(0.025*height);j<startY+(int)ceil(0.025*height);j++)
					{
					if(image[j*width+i]>maxVal)
						{
						//we need to check that the max position is not in 
						//same zone as a previous max.
						if(!isInBadZone(i,j))
							{
							maxVal=image[j*width+i];
							maxX=i;
							maxY=j;
							}
						}
					}
				}
			if(maxVal > 5*average)
				{
				maxXs.push_back(maxX);
				maxYs.push_back(maxY);
				printf("Ignoring 20x20 box around (%d,%d)\n",maxX,maxY);
				}
			}
		
		//Before doing anything we need to check that our start point is not in
		//a risky zone.	
		if(isInBadZone(startX,startY))
			{
			returnval[0]=-1;
			returnval[1]=-1;
			returnval[2]=-1;
			return returnval;
			}
		
		//To determine a circle we need three not related points. In order
		//to find this point we choose a start angle 'a' at random. Then we shoot
		//3 rays in the directions pointed by a, a+120 and a+240. We calculate
		//the fitting circle and we repeat 12 times by shifting the start angle
		//by 30 degree each time. We always keep the larger fitting circle. This
		//should prevent from being block by an artifact within backstop area.
		for(int i=15;i<375;i+=30)
			{
			double loccos,locsin;
			float myfitX,myfitY,myfitR;
			int* angle;
				
			angle=new int[3];
			angle[0]=i;
			angle[1]=i+121;
			angle[2]=i+242;
				
			for(int j=0;j<3;j++)
				{
				loccos=cos(angle[j]*(M_PI/180.0));
				locsin=sin(angle[j]*(M_PI/180.0));
				if(rayOnTileJoint(startX,startY,loccos,locsin))
					{
					angle[j]+=10;
					loccos=cos(angle[j]*(M_PI/180.0));
					locsin=sin(angle[j]*(M_PI/180.0));
					}					
				ind=ray(startX,startY,loccos,locsin,magic);
				coords[j*2]=(int) ceil(startX+ind*loccos);
				coords[j*2+1]=(int) ceil(startY+ind*locsin);
				}
				
			//We need to check that our directions are sensible, i.e. that we
			//did not accidently fell onto the arm of the backstop. if one of the
			//ray grew more than 1.5 as far as the other two then we guess that 
			//is the case and we try all three axis in reversed direction we should
			//guarantee that we do not fall on the armstop again.
			
			distances[0]=sqrt((double)((coords[0]-startX)*(coords[0]-startX)+
							 (coords[1]-startY)*(coords[1]-startY)));
			distances[1]=sqrt((double)((coords[2]-startX)*(coords[2]-startX)+
							 (coords[3]-startY)*(coords[3]-startY)));
			distances[2]=sqrt((double)((coords[4]-startX)*(coords[4]-startX)+
							 (coords[5]-startY)*(coords[5]-startY)));
			if((distances[0] > 1.5*distances[1] && distances[0] > 1.5*distances[2]) ||
			   (distances[1] > 1.5*distances[0] && distances[1] > 1.5*distances[2]) ||
			   (distances[2] > 1.5*distances[0] && distances[2] > 1.5*distances[1]))
				{
				for(int j=0;j<3;j++)
					{
					loccos=-cos(angle[j]*(M_PI/180.0));
					locsin=-sin(angle[j]*(M_PI/180.0));						
					ind=ray(startX,startY,loccos,locsin,magic);
					coords[j*2]=(int) ceil(startX+(ind-1)*loccos);
					coords[j*2+1]=(int) ceil(startY+(ind-1)*locsin);
					}
				}
				
			if(fitCircle(coords[0],coords[1],coords[2],coords[3],coords[4],coords[5],
					  myfitX,myfitY,myfitR)==0)
				{
				fitX+=(int) myfitX;
				fitY+=(int) myfitY;
				if(myfitR>fitR)
					fitR=myfitR;
				ncircles++;
				}
			}
			
		if(fitR < 0.1*width && ncircles > 1)
			{
			endX=(int) floor(fitX/(ncircles*1.0));
			endY=(int) floor(fitY/(ncircles*1.0));
			endR=fitR;
			}

		//Now we want to see if we can detect some shadow or extra lower grey
		//zones that we want to mask.
		ncircles=0;
		fitX=0;
		fitY=0;
		fitR=0;
		for(int i=0;i<360;i+=30)
			{
			int restartX,restartY;
			double loccos,locsin;
			float myfitX,myfitY,myfitR;
			int* angle;
				
			angle=new int[3];
			angle[0]=i;
			angle[1]=i+120;
			angle[2]=i+240;
				
			for(int j=0;j<3;j++)
				{
				loccos=cos(angle[j]*(M_PI/180.0));
				locsin=sin(angle[j]*(M_PI/180.0));
				restartX=(int) ceil(endX+(endR*1.01)*loccos);
				restartY=(int) ceil(endY+(endR*1.01)*locsin);
				if(rayOnTileJoint(restartX,restartY,loccos,locsin))
					{
					angle[j]+=10;
					loccos=cos(angle[j]*(M_PI/180.0));
					locsin=sin(angle[j]*(M_PI/180.0));
					restartX=(int) ceil(endX+(endR*1.01)*loccos);
					restartY=(int) ceil(endY+(endR*1.01)*locsin);
					}					
				ind=ray(restartX,restartY,loccos,locsin,magic*0.75);
				coords[j*2]=(int) ceil(restartX+ind*loccos);
				coords[j*2+1]=(int) ceil(restartY+ind*locsin);
				}
				
			distances[0]=sqrt((coords[0]-endX)*(coords[0]-endX)+
							 (coords[1]-endY)*(coords[1]-endY));
			distances[1]=sqrt((coords[2]-endX)*(coords[2]-endX)+
							 (coords[3]-endY)*(coords[3]-endY));
			distances[2]=sqrt((coords[4]-endX)*(coords[4]-endX)+
							 (coords[5]-endY)*(coords[5]-endY));
			if((distances[0] > 1.5*distances[1] && distances[0] > 1.5*distances[2]) ||
			   (distances[1] > 1.5*distances[0] && distances[1] > 1.5*distances[2]) ||
			   (distances[2] > 1.5*distances[0] && distances[2] > 1.5*distances[1]))
				{
				for(int j=0;j<3;j++)
					{
					loccos=-cos(angle[j]*(M_PI/180.0));
					locsin=-sin(angle[j]*(M_PI/180.0));			
					restartX=(int) ceil(endX+(endR*1.01)*loccos);
					restartY=(int) ceil(endY+(endR*1.01)*locsin);						
					ind=ray(restartX,restartY,loccos,locsin,magic*0.75);
					coords[j*2]=(int) ceil(restartX+(ind-1)*loccos);
					coords[j*2+1]=(int) ceil(restartY+(ind-1)*locsin);
					}
				}
			distances[0]=sqrt((coords[0]-endX)*(coords[0]-endX)+
							 (coords[1]-endY)*(coords[1]-endY));
			distances[1]=sqrt((coords[2]-endX)*(coords[2]-endX)+
							 (coords[3]-endY)*(coords[3]-endY));
			distances[2]=sqrt((coords[4]-endX)*(coords[4]-endX)+
							 (coords[5]-endY)*(coords[5]-endY));
			if(!((distances[0] > 1.5*distances[1] && distances[0] > 1.5*distances[2]) ||
			   (distances[1] > 1.5*distances[0] && distances[1] > 1.5*distances[2]) ||
			   (distances[2] > 1.5*distances[0] && distances[2] > 1.5*distances[1])))
				{
				if(fitCircle(coords[0],coords[1],coords[2],coords[3],coords[4],coords[5],
						  myfitX,myfitY,myfitR)==0)
					{
					fitX+=(int) myfitX;
					fitY+=(int) myfitY;
					if(myfitR>fitR)
						fitR=myfitR;
					ncircles++;
					}
				}
			}	
		
		if((fitR < 0.1*width && ncircles > 1 ) &&
		   (fabs(fitX/(ncircles*1.0)- endX) < 0.02*width &&
			fabs(fitY/(ncircles*1.0)- endY) < 0.02*height))
			{
			endX=(int) floor(fitX/(ncircles*1.0));
			endY=(int) floor(fitY/(ncircles*1.0));
			endR=fitR;
			}

		//ok we now got our circle, next we try to jiggle it around to see if we
		//can improve it.
		variance=0;
		ncircles=0;
		oldvariance=10000;
		fitX=endX;
		fitY=endY;
		fitR=endR;
		while(variance<oldvariance)
			{
			//if one of the move was successfull then we try to grow the circle
			if(trygrow)
				{
				average=averageDisc(fitX,fitY,fitR);
				variance=varianceDisc(fitX,fitY,fitR);
				while(sqrt(variance) < magic*average)
					{
					average=averageDisc(fitX,fitY,fitR);
					variance=varianceDisc(fitX,fitY,fitR);
					fitR+=0.25f;
					//If we grow to much (which might happen in case the backstop
					//area on the image is not round enough, we stop because we 
					//might grow indefinitely.
					if(fitR-endR>0.05*width)
						{
						average=-1;
						oldvariance=-1;
						}
					}
				if(average!=-1)
					oldvariance=variance;
				}
			trygrow=false;
			//attempt one, move backward on the line
			variance=varianceDisc(fitX-1,fitY,fitR);
			if(variance > oldvariance)
				{
				//attempt two, move forward on the line
				variance=varianceDisc(fitX+1,fitY,fitR);	
				if(variance > oldvariance)
					{
					//attempt three, move backward on the column
					variance=varianceDisc(fitX,fitY-1,fitR);
					if(variance > oldvariance)
						{
						//attempt four, move forward on the column
						variance=varianceDisc(fitX,fitY+1,fitR);
						if(variance <= oldvariance)
							{
							fitY+=1;
							trygrow=true;
							}
						}
					else
						{
						fitY-=1;
						trygrow=true;
						}
					}
				else
					{
					fitX+=1;
					trygrow=true;
					}
				}
			else
				{
				fitX-=1;
				trygrow=true;
				}
			}
		if(fabs(fitX-endX)>0.05*width || fabs(fitY-endY)>0.05*height ||
		   (startX-fitX)*(startX-fitX)+(startY-fitY)*(startY-fitY) > fitR*fitR)
			{
			returnval[0]=endX;
			returnval[1]=endY;
			returnval[2]=endR;
			}
		else
			{
			returnval[0]=fitX;
			returnval[1]=fitY;
			returnval[2]=fitR;
			}
		return returnval;
		}
	
	int* DiffractionImage::cleanEdgeVector(int edgeX,int edgeY)
		{
		int* result;
		int angle;
		double rad;
		bool isPlate;
		result=new int[4];
		isPlate=circleIsBlank(width/2,height/2,width/2.0);
		if(isPlate)
			{			
			rad=width/2.0;
			while(circleIsBlank(width/2,height/2,rad))
				rad--;
			angle=(int) floor(180*atan((edgeY-height/2)/(1.0*(edgeX-width/2)))/M_PI);
			angle=(360+angle) %360;
			}
		if(edgeX>=width-1)
			{
			result[0]=width-1;
			result[2]=0;
			result[3]=1;
			//need to adjust limit to avoid blank borders.
			if(isPlate)
				{
				result[0]=(int) floor(width/2+rad*cos(angle*M_PI/180.0));
				if(angle >315)
					result[1]=(int) floor(height/2+rad*sin(angle*M_PI/180.0));
				else
					result[1]=(int) ceil(height/2+rad*sin(angle*M_PI/180.0));
				}
			else
				{
				while(lineIsBlank(result[0],true))
					result[0]--;
				result[1]=edgeY;
				}
			}
		else
			{
			if(edgeX<=0)
				{
				result[0]=0;
				result[2]=0;
				result[3]=-1;
				//need to adjust limit to avoid blank borders.
				if(isPlate)
					{
					result[0]=(int) ceil(width/2+rad*cos(angle*M_PI/180.0));
					if(angle >180)
						result[1]=(int) floor(height/2+rad*sin(angle*M_PI/180.0));
					else
						result[1]=(int) ceil(height/2+rad*sin(angle*M_PI/180.0));
					}
				else
					{
					while(lineIsBlank(result[0],true))
						result[0]++;
					result[1]=edgeY;
					}
				}
			else
				{
				if(edgeY>=height-1)
					{
					result[1]=height-1;
					result[2]=-1;
					result[3]=0;
					if(isPlate)
						{
						result[1]=(int) floor(height/2+rad*sin(angle*M_PI/180.0));
						if(angle>90)
							result[0]=(int) floor(width/2+rad*cos(angle*M_PI/180.0));
						else
							result[0]=(int) ceil(width/2+rad*cos(angle*M_PI/180.0));
						}
					else
						{
						//need to adjust limit to avoid blank borders.
						while(lineIsBlank(result[1],false))
							result[1]--;
						result[0]=edgeX;
						}
					}
				else
					{
					result[1]=0;
					result[2]=1;
					result[3]=0;
					if(isPlate)
						{
						result[1]=(int) ceil(height/2+rad*sin(angle*M_PI/180.0));
						if(angle<270)
							result[0]=(int) floor(width/2+rad*cos(angle*M_PI/180.0));
						else
							result[0]=(int) ceil(width/2+rad*cos(angle*M_PI/180.0));
						}
					else
						{
						//need to adjust limit to avoid blank borders.
						while(lineIsBlank(result[1],false))
							result[1]++;
						result[0]=edgeX;
						}
					}
				}
			}
		return result;
		}
		
	int* DiffractionImage::lowestPixelAround(int fromX, int fromY,int vecX, int vecY)
		{	
		int lowestX,lowestY,lowestVal,initVal;
		int* toreturn;
		
		toreturn= new int[2];
		//next we scan +-100 pixels along the scan direction to choose
		//the pixel with the lowest count as start point. We expect that
		//this will position us more or less in the middle of the arm.
		initVal=image[fromY*width+fromX];
		lowestVal=initVal;
		lowestX=fromX;
		lowestY=fromY;
		for(int j=-50;j<50;j++)
			{
			int scanInd;
			scanInd=((fromY+j*vecY)*width)+(fromX+j*vecX);
			//we want a pixel count bigger than 1 to avoid catching a 
			//tile joint and dumb statistics.
			if((image[scanInd]<lowestVal && initVal-image[scanInd] >= 5) &&
				image[scanInd] > 0)
				{
				lowestX=fromX+j*vecX;
				lowestY=fromY+j*vecY;
				lowestVal=image[scanInd];
				}
			}
		toreturn[0]=lowestX;
		toreturn[1]=lowestY;
		return toreturn;
		}
		
	int* DiffractionImage::simpleArmFit(int* from, int* to,float angle,float bckstopRad)
		{	
		int nsteps,vecX,vecY;
		int* simpleFit;
		int* start;
		int* edge;
		int* clean;
		double threshold,lastdist,arc,dist,rad;
		simpleFit=new int[8];
		
		threshold=0.5;
		clean=cleanEdgeVector(to[0],to[1]);
		vecX=-clean[2];
		vecY=-clean[3];
		lastdist=bckstopRad*0.5;
		dist=1000;
		start=lowestPixelAround(from[0],from[1],vecX,vecY);		
		edge=lowestPixelAround(clean[0],clean[1],vecX,vecY);
		while(3*lastdist < dist)
			{
			vecY=-vecY;
			vecX=-vecX;
			nsteps=ray(start[0],start[1],vecX,vecY,threshold);	
			simpleFit[0]=start[0]+nsteps*vecX;
			simpleFit[1]=start[1]+nsteps*vecY;
			vecY=-vecY;
			vecX=-vecX;
			nsteps=ray(start[0],start[1],vecX,vecY,threshold);	
			simpleFit[2]=start[0]+nsteps*vecX;
			simpleFit[3]=start[1]+nsteps*vecY;
			dist=abs(simpleFit[0]-simpleFit[2])+abs(simpleFit[1]-simpleFit[3]);
			threshold/=1.25;
			}
		lastdist=abs(simpleFit[0]-simpleFit[2])+abs(simpleFit[1]-simpleFit[3]);	
		dist=1000;
		
		threshold=0.45;
		while(3*lastdist < dist)
			{
			//Check if the image is a plate, dealing with the edge is more
			//sensitive.
			if(circleIsBlank(width/2,height/2,width/2.0))
				{				
				rad=width/2.0;
				while(circleIsBlank(width/2,height/2,rad))
					rad--;
				//Firstly increasing angle
				arc=(angle+arcRay(angle,rad,true,threshold))*M_PI/180.0;			
				if(width/2+rad*cos(arc)-fabs(width/2+rad*cos(arc)) >0.5)
					simpleFit[6]=(int) ceil(width/2+rad*cos(arc));
				else
					simpleFit[6]=(int) floor(width/2+rad*cos(arc));
				if(height/2+rad*sin(arc)-fabs(height/2+rad*sin(arc)) >0.5)
					simpleFit[7]=(int) ceil(height/2+rad*sin(arc));
				else
					simpleFit[7]=(int) floor(height/2+rad*sin(arc));
				//Secondly Going the other way
				arc=(angle+arcRay(angle,rad,false,threshold))*M_PI/180.0;	
				if(width/2+rad*cos(arc)-fabs(width/2+rad*cos(arc)) >0.5)
					simpleFit[4]=(int) ceil(width/2+rad*cos(arc));
				else
					simpleFit[4]=(int) floor(width/2+rad*cos(arc));
				if(height/2+rad*sin(arc)-fabs(height/2+rad*sin(arc)) >0.5)
					simpleFit[5]=(int) ceil(height/2+rad*sin(arc));
				else
					simpleFit[5]=(int) floor(height/2+rad*sin(arc));
				dist=sqrt((double)((simpleFit[7]-simpleFit[5])*(simpleFit[7]-simpleFit[5])+
						           (simpleFit[6]-simpleFit[4])*(simpleFit[6]-simpleFit[4])));
				}
			else
				{
				vecY=-vecY;
				vecX=-vecX;
				nsteps=ray(edge[0],edge[1],vecX,vecY,threshold);	
				simpleFit[6]=edge[0]+nsteps*vecX;
				simpleFit[7]=edge[1]+nsteps*vecY;
				vecY=-vecY;
				vecX=-vecX;
				nsteps=ray(edge[0],edge[1],vecX,vecY,threshold);	
				simpleFit[4]=edge[0]+nsteps*vecX;
				simpleFit[5]=edge[1]+nsteps*vecY;
				dist=abs(simpleFit[7]-simpleFit[5])+abs(simpleFit[6]-simpleFit[4]);
				}
			threshold/=1.25;
			}
		return simpleFit;
		}
	
	int* DiffractionImage::complexArmFit(int* from, int* to,float angle,float bckstopRad)
		{
		int nsteps,vecX,vecY,lastcoord,laststored,lasti,limit;
		int* complexFit;
		int* side1;
		int* side2;
		int* clean;
		double threshold1,threshold2,dirX,dirY,normDir,rad,aver;
		double dist1,dist2,lastdist1,lastdist2;
		bool isPlate;
		
		complexFit=new int[165];
		side1=new int[82];
		side2=new int[82];
		clean=cleanEdgeVector(to[0],to[1]);
		vecX=clean[2];
		vecY=clean[3];
		isPlate=circleIsBlank(width/2,height/2,width/2.0);
		if(isPlate)
			limit=39;
		else 
			limit=40;	
		//printf("border: %d %d\n",clean[0],clean[1]);
			
		dirX=clean[0]-from[0];
		dirY=clean[1]-from[1];
		normDir=sqrt((dirX*dirX)+(dirY*dirY));
		dirX/=normDir;
		dirY/=normDir;
		lastcoord=0;
		lastdist1=bckstopRad*0.25;
		lastdist2=bckstopRad*0.25;
		threshold1=0.5;
		threshold2=0.5;
		lasti=0;
			
		for(int i=0;i<=limit;i++)
			{
			int* newStart;
			if(lasti!=i)
				{
				threshold1=0.5;
				threshold2=0.5;
				}
			lasti=i;
			//we start from the position on the ray that intersect in the
			//portion we are looking for.
			if(i<40) 
				newStart=lowestPixelAround((int) floor(from[0]+(i*0.025*normDir*dirX)),
										   (int) floor(from[1]+(i*0.025*normDir*dirY)),
										   vecX,vecY);
			else
				newStart=lowestPixelAround(clean[0],clean[1],vecX,vecY);
			//scan for point
			if (newStart[1]==2999)
				newStart[1]=2998;
			nsteps=ray(newStart[0],newStart[1],vecX,vecY,threshold1);
			side1[2*i]=newStart[0]+nsteps*vecX;
			side1[(2*i)+1]=newStart[1]+nsteps*vecY;
			dist1=abs(side1[2*i]-newStart[0])+abs(side1[(2*i)+1]-newStart[1]);
			//now we reverse the direction of search and take another point
			vecY=-vecY;
			vecX=-vecX;
			nsteps=ray(newStart[0],newStart[1],vecX,vecY,threshold2);
			side2[2*i]=newStart[0]+nsteps*vecX;
			side2[(2*i)+1]=newStart[1]+nsteps*vecY;
			dist2=abs(side2[2*i]-newStart[0])+abs(side2[(2*i)+1]-newStart[1]);		
			vecY=-vecY;
			vecX=-vecX;			
			
			//look the last points gap, if it is significantly bigger this means
			//we are falling in a low background area, therefore we need to 
			//be more strict, lower the threshold and start this iteration again
			/*aver=rayAverage(newStart[0],newStart[1],vecX,vecY,nsteps+1);*/
			if(false)
				{
				printf("start: %d %d ,dir: %d %d\n",newStart[0],newStart[1],vecX,vecY);
				printf("threshold1:%.5f,threshold2:%.5f,average:%.5f\n",threshold1,threshold2,aver);
				printf("lastdist1:%.5f,lastdist2:%.5f,dist1:%.5f,dist2:%.5f\n",lastdist1,lastdist2,dist1,dist2);
				}
			
			//This sections check the points we have generated, we sometimes need
			//to adjust the threshold to take variation in the image background 
			//into account.
			if(dist1+dist2 > 3*(lastdist1+lastdist2))
				{
				threshold1/=1.25;
				threshold2/=1.25;
				i--;
				}
			else
				{
				if(dist1 > 3*lastdist1)
					{
					threshold1/=1.25;
					i--;
					}
				else
					{	
					if(dist2 > 3*lastdist2)
						{
						threshold2/=1.25;
						i--;
						}
					else
						{	
						if(3*(dist1+dist2) < lastdist1+lastdist2)
							{
							threshold1=threshold1*1.75;
							threshold2=threshold2*1.75;
							i--;
							}
						else
							{
							if(3*dist1 < lastdist1)
								{
								threshold1=threshold1*1.75;
								i--;
								}
							else
								{
								if(3*dist2 < lastdist2)
									{
									threshold2=threshold2*1.75;
									i--;
									}
								else
									{
									lastdist1=dist1;
									lastdist2=dist2;
									}
								}
							}
						}
					}
				}
			delete newStart;
			}	
		
		threshold1=0.5;
		threshold2=0.5;			
		if(isPlate)
			{
			double arc;
			int* newStart;
			newStart=lowestPixelAround(clean[0],clean[1],vecX,vecY);
			rad=width/2.0;
			while(circleIsBlank(width/2,height/2,rad))
				rad--;
			//Firstly increasing angle
			arc=(angle+arcRay(angle,rad,true,threshold1))*M_PI/180.0;			
			if(width/2+rad*cos(arc)-fabs(width/2+rad*cos(arc)) >0.5)
				side1[80]=(int) ceil(width/2+rad*cos(arc));
			else
				side1[80]=(int) floor(width/2+rad*cos(arc));
			if(height/2+rad*sin(arc)-fabs(height/2+rad*sin(arc)) >0.5)
				side1[81]=(int) ceil(height/2+rad*sin(arc));
			else
				side1[81]=(int) floor(height/2+rad*sin(arc));
			//Secondly Going the other way
			arc=(angle+arcRay(angle,rad,false,threshold2))*M_PI/180.0;	
			if(width/2+rad*cos(arc)-fabs(width/2+rad*cos(arc)) >0.5)
				side2[80]=(int) ceil(width/2+rad*cos(arc));
			else
				side2[80]=(int) floor(width/2+rad*cos(arc));
			if(height/2+rad*sin(arc)-fabs(height/2+rad*sin(arc)) >0.5)
				side2[81]=(int) ceil(height/2+rad*sin(arc));
			else
				side2[81]=(int) floor(height/2+rad*sin(arc));
			/*dist=sqrt((double)(((side2[78]-side1[78])*(side2[78]-side1[78]))+
							   ((side2[79]-side1[79])*(side2[79]-side1[79]))));*/
			delete newStart;
			}

		/*printf("last points %d %d, %d %d?\n",lastThree[4*lastcoord],
			   lastThree[4*lastcoord+1],lastThree[4*lastcoord+2],
			   lastThree[4*lastcoord+3]);*/
		int* vectors= new int[8];
		complexFit[0]=side1[0];
		complexFit[1]=side1[1];
		complexFit[2]=side2[0];
		complexFit[3]=side2[1];
		laststored=1;			
		for(int i=1;i<40;i++)
			{
			double dot1, dot2;
			double norm1,norm2;
			//we have 6 points which could mean two quads, we check to see
			//if this can be reduced to a single quads, given that the edge 
			//have a small angle between them.
			vectors[0]=side1[(2*i)]-complexFit[(laststored-1)*4];
			vectors[1]=side1[(2*i)+1]-complexFit[(laststored-1)*4+1];
			vectors[2]=side2[(2*i)]-complexFit[(laststored-1)*4+2];
			vectors[3]=side2[(2*i)+1]-complexFit[(laststored-1)*4+3];
			vectors[4]=side1[(2*(i+1))]-side1[(2*i)];
			vectors[5]=side1[(2*(i+1))+1]-side1[(2*i)+1];
			vectors[6]=side2[(2*(i+1))]-side2[(2*i)];
			vectors[7]=side2[(2*(i+1))+1]-side2[(2*i)+1];
			//compare the dot products
			dot1=vectors[0]*vectors[4]+vectors[1]*vectors[5];
			dot2=vectors[2]*vectors[6]+vectors[3]*vectors[7];
			norm1=sqrt((double)(vectors[0]*vectors[0]+vectors[1]*vectors[1]))*
				  sqrt((double)(vectors[4]*vectors[4]+vectors[5]*vectors[5]));
			norm2=sqrt((double)(vectors[2]*vectors[2]+vectors[3]*vectors[3]))*
				  sqrt((double)(vectors[6]*vectors[6]+vectors[7]*vectors[7]));
			/*printf("dot1: %.3f dot2: %.3f norm1: %.3f norm2: %.3f, cos1:%.3f, co22:%.3f\n",
			       dot1,dot2,norm1,norm2,dot1/norm1,dot2/norm2);*/
			if(fabs(dot1/norm1)<0.975 || fabs(dot2/norm2)<0.975)
				{
				//the edges are not flat enough store coordinates and continue
				complexFit[laststored*4]=side1[2*i];
				complexFit[laststored*4+1]=side1[(2*i)+1];
				complexFit[laststored*4+2]=side2[2*i];
				complexFit[laststored*4+3]=side2[(2*i)+1];
				laststored++;
				}
			}
		complexFit[laststored*4]=side1[80];
		complexFit[laststored*4+1]=side1[81];
		complexFit[laststored*4+2]=side2[80];
		complexFit[laststored*4+3]=side2[81];			
		complexFit[laststored*4+4]=-1;
		delete vectors;
		delete side1;
		delete side2;
		return complexFit;
		}
		
	//This one is needed to avoid shooting ray along CCD tiles junction as pixel
	//count there is null and therefore mess up with averages and variances.
	bool DiffractionImage::rayOnTileJoint(int fromX, int fromY,double vecX, double vecY)
		{
		bool returnval;
		list<int> joints;
		list<int>::iterator j;
			
		if(width%2 == 0 && height %2==0)
			{
			if(lineIsBlank(width/2,true) || lineIsBlank(width/2,false))
				{
				joints.push_back(width/2);
				}
			}
		
		if(width%3 == 0 && height %3==0)
			{
			if(lineIsBlank(width/3,true) || lineIsBlank(width/3,false))
				{
				joints.push_back(width/3);
				joints.push_back(2*width/3);
				}
			}
			
		if(manufact.compare("Dectris") == 0)
			{
			//some fun to foresee with the countless tiles of a pilatus image
			}			
		
		returnval=false;
			
		for(j=joints.begin();j!=joints.end();j++)
			{
			if(fabs(vecY) > 0.998)
				{
				if(fabs(vecY)==1)
					{
					if(abs(fromX-(*j))<3)
						returnval=true;
					else
						returnval=false;
					}
				else
					{
					int inter;
					double t=((*j)-fromX)/vecX;
					inter=(int) floor(fromY+t*vecY);
					if(inter<0 || inter > height-1)
						returnval=false;
					else
						returnval=true;
					}
				}
			else
				{
				if(fabs(vecX)>0.998)
					{
					if(fabs(vecX)==1)
						{
						if(abs(fromY-(*j))<3)
							returnval=true;
						else
							returnval=false;
						}
					else
						{
						int inter;
						double t=((*j)-fromY)/vecY;
						inter=(int) floor(fromX+t*vecX);
						if(inter<0 || inter > width-1)
							returnval=false;
						else
							returnval=true;
						}
					}
				else
					returnval=false;
				}
			}
		return returnval;
		}
	
	//check if a line contains too many blank pixels.
	bool DiffractionImage::lineIsBlank(int start,bool horizontal)
		{
		int fromX, fromY,limit,nblank,dirX,dirY;
		if(horizontal)
			{
			fromX=0;
			fromY=start;
			dirX=1;
			dirY=0;	
			limit=width;
			}
		else
			{
			fromY=0;
			fromX=start;
			dirX=0;
			dirY=1;	
			limit=height;
			}
		nblank=0;
		for(int i=0;i<limit;i++)
			{
			if(image[(fromX+i*dirX)+(fromY+i*dirY)*width]==0)
				nblank++;
			}
		//if more than 15 percent of the line is blank than we reject it.
		return (nblank > 0.15*width);
		}
		
	//check if a circle contains too many white pixels. This is used in the case
	//the image is in the form of an image plate.
	bool DiffractionImage::circleIsBlank(int centX,int centY,float radius)
		{
		int nblank,circleX,circleY;
		nblank=0;
		for(int i=0;i<360;i++)
			{
			circleX=(int) floor(centX+radius*cos(i*M_PI/180.0));
			if(fabs(circleX-(centX+radius*cos(i*M_PI/180.0))) > 0.5);
				circleX=(int) ceil(centX+radius*cos(i*M_PI/180.0));	
			circleY=(int) floor(centY+radius*sin(i*M_PI/180.0));
			if(fabs(circleY-(centY+radius*sin(i*M_PI/180.0))) > 0.5);
				circleY=(int) ceil(centY+radius*sin(i*M_PI/180.0));	
			if((circleX >= 0 && circleX <= width-1) &&
			   (circleY >= 0 && circleY <= height-1))
				{
				if(image[circleX+(circleY*width)]==0)
					nblank++;
				}
			}
		//if more than 5 percent of the line is blank than we reject it.
		return (nblank > 120);
		}
		
	//calculate the mean along one arc		
	float DiffractionImage::arcAverage(float rad,float startAng,float endAng)
		{
		int index,lastindex,npix,sum,pixX,pixY;
		float average,step,start,end;
		
		sum=0;
		lastindex=-1;
		npix=0;
		if(endAng>startAng)
			{
			start=startAng*M_PI/180.0;
			end=endAng*M_PI/180.0;
			}
		else
			{
			start=endAng*M_PI/180.0;
			end=startAng*M_PI/180.0;
			}

		step=atan(1.0/(rad*1.0));
		
		for(float i=start;i<=end;i+=step)
			{
			if(width/2+rad*cos(i)-fabs(width/2+rad*cos(i)) > 0.5)
				pixX=(int) ceil(width/2+rad*cos(i));
			else
				pixX=(int) floor(width/2+rad*cos(i));
			if(height/2+rad*sin(i)-fabs(height/2+rad*sin(i)) > 0.5)
				pixY=(int) ceil(height/2+rad*sin(i));
			else
				pixY=(int) floor(height/2+rad*sin(i));
			index=pixY*width+pixX;
			if(index !=lastindex && image[index]>0)
				{
				sum+=image[index];
				lastindex=index;
				npix++;
				}
			}
			
		average=sum/(1.0f*npix);
		return average;
		}
	
	//calculate the variance along one arc		
	float DiffractionImage::arcVariance(float rad,float startAng,float endAng)
		{
		int index,lastindex,npix,pixX,pixY;
		float step,sum,sumsq,start,end,variance;
		
		sum=0;
		sumsq=0;
		lastindex=-1;
		npix=0;
		if(endAng>startAng)
			{
			start=startAng*M_PI/180.0;
			end=endAng*M_PI/180.0;
			}
		else
			{
			start=endAng*M_PI/180.0;
			end=startAng*M_PI/180.0;
			}

		step=atan(1.0/(rad*1.0));
			
		for(float i=start;i<=end;i+=step)
			{
			if(width/2+rad*cos(i)-fabs(width/2+rad*cos(i)) > 0.5)
				pixX=(int) ceil(width/2+rad*cos(i));
			else
				pixX=(int) floor(width/2+rad*cos(i));
			if(height/2+rad*sin(i)-fabs(height/2+rad*sin(i)) > 0.5)
				pixY=(int) ceil(height/2+rad*sin(i));
			else
				pixY=(int) floor(height/2+rad*sin(i));
			index=pixY*width+pixX;
			if(index !=lastindex && image[index]>0)
				{
				sum+=image[index];
				sumsq+=(image[index]*image[index]);
				lastindex=index;
				npix++;
				}
			}
		variance=(sumsq/(1.0f*npix))-((sum/(1.0f*npix))*(sum/(1.0f*npix)));
		return variance;
		}
		
	//Shoots an arc ray and return the stop angle in Radian
	float DiffractionImage::arcRay(float fromAng, float radius, bool plus, double thresh)
		{
		int lastdist,returnval;
		float average,variance,arc,stepDeg;
		
		if(plus)
			stepDeg=180*(atan(1.0/(radius*1.0))/M_PI);
		else
			stepDeg=-180*(atan(1.0/(radius*1.0))/M_PI);
		
		arc=stepDeg;
		
		//Threshold is too low.
		if(thresh < 0.1)
			return -1;
		
		//We use the magic comparison of the variance and threshold 
		//times the average
		average=arcAverage(radius,fromAng,fromAng+arc);
		variance=arcVariance(radius,fromAng,fromAng+arc);
		while( sqrt(variance) < thresh*average && arc < 25)
			{
			average=arcAverage(radius,fromAng,fromAng+arc);
			variance=arcVariance(radius,fromAng,fromAng+arc);
			arc+=stepDeg;
			}
		if(arc > 25)
			{
			return -1;
			}
		else
			return arc-stepDeg;
		}	
		
	//calculate the mean along one ray		
	float DiffractionImage::rayAverage(int fromX,int fromY,double vecX,
		                               double vecY, int steps)
		{
		int index,lastindex,npix,sum;
		float average;
		//First get average value along the ray
		sum=0;
		lastindex=-1;
		npix=0;
		for(int i=0;i<steps;i++)
			{
			int ivecY,ivecX;
			if((i*vecX)-floor(i*vecX)>0.5)
				ivecX=(int) ceil(i*vecX);
			else
				ivecX=(int) floor(i*vecX);
			if((i*vecY)-floor(i*vecY)>0.5)
				ivecY=(int) ceil(i*vecY);
			else
				ivecY=(int) floor(i*vecY);
			if((fromY+ivecY <0 || fromX+ivecX < 0) ||
			   (fromY+ivecY > height-1 || fromX+ivecX>width-1) )
				{
				printf("coordinates: %d %d\n",fromX+ivecX,fromY+ivecY);
				throw (DiffractionImageException("rayVariance: Unexpected step out of image"));
				}
			index=((fromY+ivecY)*width)+(fromX+ivecX);
			if(index!=lastindex && image[index] > 0)
				{
				//If pixel is an acknowledged risky zone around the "beam centre" than
				//we do not include it in the statistics
				if(!isInBadZone(fromX+ivecX,fromY+ivecY))
					{
					sum+=image[index];
					npix++;
					}
				lastindex=index;
				}
			}
		if(npix > 0)
			average=sum/(1.0f*npix);
		else
			average=-1;
		return average;
		}
	
	//calculate the variance along one ray		
	float DiffractionImage::rayVariance(int fromX,int fromY,double vecX,
		                               double vecY, int steps)
		{
		int index,lastindex,npix;
		float variance,sum,sumsq;
		
		sum=0;
		sumsq=0;
		lastindex=-1;
		npix=0;
		for(int i=0;i<steps;i++)
			{
			int ivecY,ivecX;
			if((i*vecX)-floor(i*vecX)>0.5)
				ivecX=(int) ceil(i*vecX);
			else
				ivecX=(int) floor(i*vecX);
			if((i*vecY)-floor(i*vecY)>0.5)
				ivecY=(int) ceil(i*vecY);
			else
				ivecY=(int) floor(i*vecY);
			if((fromY+ivecY <0 || fromX+ivecX < 0) ||
			   (fromY+ivecY > height-1 || fromX+ivecX>width-1) )
				{
				printf("coordinates: %d %d\n",fromX+ivecX,fromY+ivecY);
				throw (DiffractionImageException("rayVariance: Unexpected step out of image"));
				}
			index=((fromY+ivecY)*width)+(fromX+ivecX);
			if(index!=lastindex && image[index] > 0)
				{
				//If pixel is an acknowledged risky zone around the "beam centre" than
				//we do not include it in the statistics
				if(!isInBadZone(fromX+ivecX,fromY+ivecY))
					{
					sum+=image[index];
					sumsq+=(image[index]*image[index]);
					npix++;
					}
				lastindex=index;
				}
			}
		if(npix > 0)
			variance=(sumsq/(1.0f*npix))-((sum/(1.0f*npix))*(sum/(1.0f*npix)));
		else
			variance=-1;
		return variance;
		}
		
	//Shoots a ray and return the distance to stop point
	int DiffractionImage::ray(int fromX,int fromY,double vecX,double vecY,double thresh)
		{
		int lastdist,nsteps,returnval;
		float average,variance;
		bool edge;
			
		nsteps=1;
		edge=false;
		
		//Threshold is too low.
		if(thresh < 0.1)
			return -1;
		
		//We use the magic comparison of the variance and threshold 
		//times the average
		average=rayAverage(fromX,fromY,vecX,vecY,nsteps);
		variance=rayVariance(fromX,fromY,vecX,vecY,nsteps);
		while(sqrt(variance) < thresh*average && ! edge)
			{
			nsteps++;
			edge=(fromX+nsteps*vecX >= width-1 || fromX+nsteps*vecX <=0) ||
				 (fromY+nsteps*vecY >= height-1 || fromY+nsteps*vecY <=0);
			if(!edge)
				{
				average=rayAverage(fromX,fromY,vecX,vecY,nsteps);
				variance=rayVariance(fromX,fromY,vecX,vecY,nsteps);
				}  
			}
		if(edge)
			{
			return nsteps;
			}
		else
			return nsteps-1;
		}
		
	float DiffractionImage::averageDisc(float cx, float cy, float r)
		{
		int sumPixVal,nPixels,cxx,cyy,rr,pixX,pixY,disktest;
			
		nPixels=0;
		sumPixVal=0;
		rr=(int) ceil(r);
		if(fabs(cx)-floor(cx)>0.5)
			cxx=(int) ceil(cx);
		else
			cxx=(int) floor(cx);
		if(fabs(cy)-floor(cy)>0.5)
			cyy=(int) ceil(cy);
		else
			cyy=(int) floor(cy);
		
		if((cxx-rr < 0 || cyy-rr <0) || (cxx+rr > width-1 || cyy+rr > height-1))
			{
			printf("centre: %.2f %.2f, radius: %f\n",cx,cy,r);
			throw (DiffractionImageException("averageDisc: Specified circle is partly out of the image"));
			}
		
		for(int i=0;i<2*rr;i++)
			{
			pixX=cxx-rr+i;
			for(int j=0;j<2*rr;j++)
				{
				pixY=cyy-rr+j;
				disktest=((pixX-cxx)*(pixX-cxx))+((pixY-cyy)*(pixY-cyy));
				if(disktest<=r*r && image[(pixY)*width+pixX] > 0)
					{
					if(!isInBadZone(pixX,pixY))
						{
						sumPixVal+=image[(pixY)*width+pixX];
						nPixels++;
						}
					}
				}
			}
		return (1.0f*sumPixVal)/(1.0f*nPixels);
		}
		
	float DiffractionImage::varianceDisc(float cx, float cy, float r)
		{
		int nPixels,cxx,cyy,rr,pixX,pixY,disktest;
		float variance,sumPixVal,sumPixValsq;
			
		nPixels=0;
		sumPixVal=0;
		sumPixValsq=0;
		rr=(int) ceil(r);
		if(fabs(cx)-floor(cx)>0.5)
			cxx=(int) ceil(cx);
		else
			cxx=(int) floor(cx);
		if(fabs(cy)-floor(cy)>0.5)
			cyy=(int) ceil(cy);
		else
			cyy=(int) floor(cy);
		
		if((cxx-rr < 0 || cyy-rr <0) || (cxx+rr > width-1 || cyy+rr > height-1))
			{
			printf("centre: %.2f %.2f, radius: %f\n",cx,cy,r);
			throw (DiffractionImageException("averageDisc: Specified circle is partly out of the image"));
			}
		
		for(int i=0;i<2*rr;i++)
			{
			pixX=cxx-rr+i;
			for(int j=0;j<2*rr;j++)
				{
				pixY=cyy-rr+j;
				disktest=((pixX-cxx)*(pixX-cxx))+((pixY-cyy)*(pixY-cyy));
				if(disktest<=r*r && image[(pixY)*width+pixX] > 0)
					{
					if(!isInBadZone(pixX,pixY))
						{
						sumPixVal+=image[(pixY)*width+pixX];
						sumPixValsq+=(image[(pixY)*width+pixX]*image[(pixY)*width+pixX]);
						nPixels++;
						}
					}
				}
			}
		variance=(1.0f*sumPixValsq)/(1.0f*nPixels);
		variance-=(((1.0f*sumPixVal)/(1.0f*nPixels))*((1.0f*sumPixVal)/(1.0f*nPixels)));
		return variance;
		}
		
	float DiffractionImage::averageQuad(int* vertices)
		{
		int sumPixVal,nPixels,pixX,pixY,dirX,dirY,dir2X,dir2Y;
		int minX,maxX,minY,maxY;
			
		maxX=-1;
		maxY=-1;
		minX=width+1;
		minY=height+1;
		nPixels=0;
		sumPixVal=0;
			
		//The four limits design a box in which we are going to scan line by
		//line to test if pixel is inside or outside the box;
		for(int i=0;i<8;i+=2)
			{
			if(vertices[i]<minX)
				minX=vertices[i];
			if(vertices[i]>maxX)
				maxX=vertices[i];
			if(vertices[i+1]<minY)
				minY=vertices[i+1];
			if(vertices[i+1]>maxY)
				maxY=vertices[i+1];
			}	
		for(int i=minX;i<=maxX;i++)
			{
			for(int j=minY;j<=maxY;j++)
				{
				if(isInsideQuad(vertices,i,j))
					{
					nPixels++;
					sumPixVal+=image[j*width+i];
					}
				}
			}
		return (1.0f*sumPixVal)/(1.0f*nPixels);
		}
		
	float DiffractionImage::varianceQuad(int* vertices)
		{
		int nPixels,pixX,pixY,dirX,dirY,dir2X,dir2Y;
		int minX,maxX,minY,maxY;
		float variance,sum,sumsq;
			
		maxX=-1;
		maxY=-1;
		minX=width+1;
		minY=height+1;
		nPixels=0;
		sum=0;
		sumsq=0;
			
		//The four limits design a box in which we are going to scan line by
		//line to test if pixel is inside or outside the box;
		for(int i=0;i<8;i+=2)
			{
			if(vertices[i]<minX)
				minX=vertices[i];
			if(vertices[i]>maxX)
				maxX=vertices[i];
			if(vertices[i+1]<minY)
				minY=vertices[i+1];
			if(vertices[i+1]>maxY)
				maxY=vertices[i+1];
			}

		for(int i=minX;i<=maxX;i++)
			{
			for(int j=minY;j<=maxY;j++)
				{
				if(isInsideQuad(vertices,i,j))
					{
					nPixels++;
					sum+=image[j*width+i];
					sumsq+=(image[j*width+i]*image[j*width+i]);
					}
				}
			}
		//printf("averageQuad: %f, varianceQuad: %f\n",average,sum/(1.0f*nPixels));
		variance=sumsq/(1.0f*nPixels)-((sum/(1.0f*nPixels))*(sum/(1.0f*nPixels)));
		return variance;
		}
		
	bool DiffractionImage::isInBadZone(int x, int y)
		{
		list<int>::iterator X=maxXs.begin();
		list<int>::iterator Y=maxYs.begin();
		for(char i=0;i<maxXs.size();i++)
			{
			if(abs((*X)-x)<10 && abs((*Y)-y)<10)
				{
				return true;
				}
			X++;
			Y++;
			}
		return false;
		}
		
	//This function should correct the coordinates to put them just out of of the
	//bad zone box.
	int* DiffractionImage::correctCoord(int x, int y)
		{
		list<int>::iterator X=maxXs.begin();
		list<int>::iterator Y=maxYs.begin();
		int* toret=new int[2];
		for(char i=0;i<maxXs.size();i++)
			{
			if(abs((*X)-x)<10 && abs((*Y)-y)<10)
				{
				if((*X)-x>0)
					toret[0]=x+(10-((*X)-x))+2;
				else
					toret[0]=x-(10-((*X)-x))-2;
				if((*Y)-y>0)
					toret[1]=y+(10-((*Y)-y))+2;
				else
					toret[1]=y-(10-((*Y)-y))-2;
				}
			X++;
			Y++;
			}
		return toret;
		}
	};
