// DiffractionImageCompare.cpp
// maintained by G.Winter
// 8th November 2004
// 
// A set of routines for performing pixel-wise calculations between two 
// images - this image and the other image.
// 
// void min(const DiffractionImage & other);
// void max(const DiffractionImage & other);
// void diff(const DiffractionImage & other);
// void sum(const DiffractionImage & other);
// 
// $Id: DiffractionImageCompare.cpp,v 1.7 2008/05/07 11:00:51 fr45 Exp $


#include "DiffractionImage.h"

namespace DI
	{
	void DiffractionImage::min(DiffractionImage* other)
		{
		unsigned int * other_image;
		if (image == NULL)
			throw (DiffractionImageException("NULL image (this)"));
		if (width != other->getWidth())
			throw (DiffractionImageException("Images different width"));
		if (height != other->getHeight())
			throw (DiffractionImageException("Images different height"));
		other_image = other->getImage();
		if (other_image == NULL)
			throw(DiffractionImageException("NULL image (other)"));
		for (int i = 0; i < width * height; i++)
			{
			if (other_image[i] < image[i])
				image[i] = other_image[i];
			}
		realImage=false;
		}

	void DiffractionImage::max(DiffractionImage* other)
		{
		unsigned int * other_image;
		if (image == NULL)
			throw (DiffractionImageException("NULL image (this)"));
		if (width != other->getWidth())
			throw (DiffractionImageException("Images different width"));
		if (height != other->getHeight())
			throw (DiffractionImageException("Images different height"));
		other_image = other->getImage();
		if (other_image == NULL)
			throw(DiffractionImageException("NULL image (other)"));
		for (int i = 0; i < width * height; i++)
			{
			if (other_image[i] > image[i])
				image[i] = other_image[i];
			}
		realImage=false;
		}

	void DiffractionImage::diff(DiffractionImage* other)
		{
		unsigned int * other_image;
		if (image == NULL)
			throw (DiffractionImageException("NULL image (this)"));
		if (width != other->getWidth())
			throw (DiffractionImageException("Images different width"));
		if (height != other->getHeight())
			throw (DiffractionImageException("Images different height"));
		other_image = other->getImage();
		if (other_image == NULL)
			throw(DiffractionImageException("NULL image (other)"));
		for (int i = 0; i < width * height; i++)
			{
			image[i]-= other_image[i];
			if(image[i]<0)
				image[i]=0;
			}
		realImage=false;
		}

	void DiffractionImage::sum(DiffractionImage* other)
		{
		unsigned int * other_image;
		if (image == NULL)
			throw (DiffractionImageException("NULL image (this)"));
		if (width != other->getWidth())
			throw (DiffractionImageException("Images different width"));
		if (height != other->getHeight())
			throw (DiffractionImageException("Images different height"));
		other_image = other->getImage();
		if (other_image == NULL)
			throw(DiffractionImageException("NULL image (other)"));
		for (int i = 0; i < width * height; i++)
			{
			image[i]+= other_image[i];
			}
		realImage=false;
		}
	};
