// DiffractionImageOxford.cpp
// Author: F. Remacle
// August 2007
// 
#include "DiffractionImage.h"


// if we are in development mode, then include things pertaining to
// assert.
#ifdef TEST_CODE
#define NDEBUG0
#include <assert.h>
#endif

using namespace std;

namespace DI 
	{
#define RADTODEG 180.0/M_PI
	void DiffractionImage::loadOxford()
		{
		int pos,shortov,longov;
		short* over1;
		int* over2;
		string dummy,dummy2;
		char* tmpstr=new char[40];
		int myfile = open(filename.c_str(), MYREAD);
		read(myfile,tmpstr,40);
		close(myfile);
		dummy=tmpstr;
		pos=dummy.find("COMPRESSION=",0)+12;
		dummy2=dummy.substr(pos,pos+3);
		bool compressed;
		if(strstr(dummy2.c_str(),"NO")!=NULL)
			compressed=false;
		else
			compressed=true;

		// read the header		
		loadOxfordHeader();    
		allocateImage();
		myfile = open(filename.c_str(), MYREAD);
		if(!compressed)
			{
			tmpstr=new char[originalHeaderLength];
			read(myfile,tmpstr,originalHeaderLength);
			read(myfile,(char *) image,height*width*4);
			}
		else
			{
			char* tmpstr2=new char[213];
			read(myfile,tmpstr,40);
			dummy=tmpstr;
			dummy.append(1,'\n');
			read(myfile,tmpstr,3);
			read(myfile,tmpstr2,213);
			dummy.append(tmpstr2);
			//firstly check if there is some overflow and keep the numbers.
			pos=dummy.find("OI=",0)+3;
			dummy2=dummy.substr(pos,dummy.find("OL=")-1);
			shortov=atoi(dummy2.c_str());
			pos=dummy.find("OL=",0)+3;
			dummy2=dummy.substr(pos,dummy.find("NHEADER=")-1);
			longov=atoi(dummy2.c_str());
			//skip the remaining part of the header
			tmpstr=new char[originalHeaderLength];
			read(myfile,tmpstr,originalHeaderLength-256);
			//Read the raw image values
			tmpstr=new char[height*width];
			read(myfile,tmpstr,height*width);
			//read short overflow if any
			if(shortov!=0)
				{
				over1=new short[shortov];
				read(myfile,(char*) over1,shortov*2);
				}
			//read long overflow if any
			if(longov!=0)
				{
				over2=new int[longov];
				read(myfile,(char*) over2,longov*4);
				}
			//Now we need to pass through the whole array to calculate the real
			//value of the image.
			int j=0;
			int k=0;

			if((unsigned char) tmpstr[0]==254)
				{
				image[0]=over1[j];
				j++;
				}
			else
				{
				if((unsigned char) tmpstr[0]==255)
					{
					image[0]=over2[k];
					k++;
					}
				else
					image[0]=tmpstr[0]-127;
				}
			for(int i=1;i<height*width;i++)
				{
				unsigned char tmp=(unsigned char) tmpstr[i];
				int tmpint;
				if(tmp==254)
					{
					tmpint=image[i-1]+over1[j];
					j++;
					}
				else
					{
					if(tmp==255)
						{
						tmpint=image[i-1]+over2[k];
						k++;
						}
					else
						tmpint=image[i-1]+tmp-127;
					}
				//Doing the dummy offset from mosflm to be consistent
				if(tmpint<0)
					tmpint=0;
				image[i]=(unsigned int) tmpint;
				}
			}
		close(myfile);
		}
    
	void DiffractionImage::loadOxfordHeader() 
		{
		int startangles[6];
		int endangles[6];
		int section_sizes[4];
		int scanaxis,pos;
		char* tmpstr;
		char* tmpstr2;
		int* tmpint;
		short* tmpshort;
		float* tmpfloat;
		double* tmpdble;
		double oX,oY;
		int myfile;
		string dummy,dummy2;
		
		tmpstr=new char[256];
		tmpstr2=new char[256];
		myfile = open(filename.c_str(), MYREAD);
		read(myfile,tmpstr,40);
		dummy=tmpstr;
		dummy.append(1,'\n');
		read(myfile,tmpstr,3);
		read(myfile,tmpstr2,213);
		dummy.append(tmpstr2);
		if(tmpstr[3]=='R') //R for 'Ruby' detector
			{
			pixelX = 0.048;
			pixelY = 0.048;
			}
		else
			{//The other detectors, Onyx and Sapphire have a pixel size of 0.06
			pixelX = 0.06;
			pixelY = 0.06;
			//except sapphire 3 who has 0.03		
			if(tmpstr[3]=='S' && tmpstr[13]=='3')
				{
				pixelX/=2.0;
				pixelY/=2.0;
				}
			}
		pos=dummy.find("NX=",0)+3;
		dummy2=dummy.substr(pos,dummy.find("NY=")-1);
		width=atoi(dummy2.c_str());
		pos=dummy.find("NY=",0)+3;
		dummy2=dummy.substr(pos,dummy.find("OI=")-1);
		height=atoi(dummy2.c_str());
			
		pos=dummy.find("NHEADER=",0)+8;
		dummy2=dummy.substr(pos,dummy.find("NG=")-1);
		section_sizes[0]=atoi(dummy2.c_str());
		pos=dummy.find("NG=",0)+3;
		dummy2=dummy.substr(pos,dummy.find("NS=")-1);
		section_sizes[1]=atoi(dummy2.c_str());
		pos=dummy.find("NS=",0)+3;
		dummy2=dummy.substr(pos,dummy.find("NK=")-1);
		section_sizes[2]=atoi(dummy2.c_str());
		pos=dummy.find("NK=",0)+3;
		dummy2=dummy.substr(pos,dummy.find_last_of("NS=")-1);
		section_sizes[3]=atoi(dummy2.c_str());
		originalHeaderLength=section_sizes[0];
			
		tmpstr=new char[section_sizes[1]+section_sizes[2]+section_sizes[3]];
		read(myfile,tmpstr,section_sizes[1]+section_sizes[2]+section_sizes[3]);
		tmpint=(int*) tmpstr;
		tmpshort=(short*) tmpstr;
		tmpdble=(double*) tmpstr;
		close(myfile);
		for(int i=0;i<26;i++)
			date.append(1,tmpstr[section_sizes[1]+section_sizes[2]+i]);

		//Now we check how the pixel are binned => adapt the pixel size
		pixelX*=tmpshort[0];
		pixelY*=tmpshort[1];
		
		exposureTime = tmpdble[(section_sizes[1]/8)+60];
		
		beamX = (float) (tmpdble[((section_sizes[1]+section_sizes[2])/8)+83]*pixelX);
		beamY = (float) (tmpdble[((section_sizes[1]+section_sizes[2])/8)+84]*pixelY);


		//Fixme, need to calculte the origin
		oX= 0;
		oY= 0;
		
		//according to the documentation image is recorded from bottom to top
		//so when being read, the bottom is the first line read and is therefore
		//at the top.
		rapidAxis[0]=1;
		rapidAxis[1]=0;
		rapidAxis[2]=0;
		slowAxis[0]=0;
		slowAxis[1]=-1;
		slowAxis[2]=0;
			
		distance = (float) tmpdble[((section_sizes[1]+section_sizes[2])/8)+89];
		wavelength = (float) tmpdble[((section_sizes[1]+section_sizes[2])/8)+73];
		
		
		serialNo="N/A";
		twoTheta=0;
		scanaxis=-1;
		for(int i=0;i<6;i++)
			{
			startangles[i]=tmpint[((section_sizes[1]+section_sizes[2])/4)+71+i];
			endangles[i]=tmpint[((section_sizes[1]+section_sizes[2])/4)+81+i];
			if(startangles[i]!=endangles[i])
				scanaxis=i;
			}			
		
		if(scanaxis!=-1)
			{
			oscStart = (float) startangles[scanaxis]*RADTODEG;
			oscEnd = (float) endangles[scanaxis]*RADTODEG;
			oscStart*=(float) tmpdble[((section_sizes[1]+section_sizes[2])/8)+46];
			oscEnd*=(float) tmpdble[((section_sizes[1]+section_sizes[2])/8)+46];
			}
			
		switch(scanaxis)
			{
			case 0:
				oscAxis = "omega";
				break;
			case 1:
				oscAxis = "theta";
				break;
			case 2:
				oscAxis = "kappa / chi";
				break;
			case 3:
				oscAxis = "phi";
				break;
			case 4:
				oscAxis = "omega'";
				break;
			case 5:
				oscAxis = "theta'";
				break;
			default:
				oscAxis = "unknown";
				break;
			}
		serialNo = "N/A";
		format="CRYSALIS";
		manufact="OXFORD-DIFFRACTION";
		}
	};
