// Profile.h
// maintained by F.Remacle
// 4th Septembre 2007
// 
// 
// $Id: Profile.h,v 1.4 2007/10/31 09:40:17 fr45 Exp $
// 

#ifndef _PROFILE_HEADER
#define _PROFILE_HEADER

#include "DiffractionImage.h"

// standard templates which I want to use

#include <list>
#include <vector>

struct ProfPeak {
  int posX;  				//x position of the peak in pixels
  int posY;  				//y position of the peak in pixels
  int indInt; 				//index of peak in the intensities array
  int distFromPrevious;		//Distance from previous Peak in angstrom
  unsigned int intensity;	//intensity of the peak
};

namespace DI
{
  class Profile
    {
    public:
#ifdef INCLUDE_PYTHON
      PyObject_HEAD;
#endif
      // constructors and destructors
      Profile(void);

      // this will create a populated peak list
      Profile(DiffractionImage * diffractionImage, int fromX, int fromY, int toX, int toY);

      ~Profile(void);

      // things to aid the Python interfacification
      void constructor(void);
      void constructor(DiffractionImage * diffractionImage, int fromX, int fromY, int toX, int toY);

      void destructor(void);

      // next the methods
      int getNPeaks(void);
	  int getLength(void);
	  unsigned int getMax(void);
      void loadProfile(DiffractionImage * diffractionImage, int fromX, int fromY, int toX, int toY);

      // an array dereference operator and the equivalent method
      ProfPeak & get(int offset);

      // this will make a complete copy of the list and return this as a
      // vector - which will be better for performing numerical things
      // with
      ProfPeak* getPeaks(void);
	  
	  unsigned int* getIntensities();
	  
    private:
      // __PeakList * _list;
	  unsigned int * intensities;
	  unsigned int intMax;
	  int * Xs;
	  int * Ys;
      list<ProfPeak> _list;
      int _length;
	  int npixels;
    };

};

#endif
